# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import datetime
import http.client
import random
from unittest import mock
import urllib
from urllib import parse as urlparse
import uuid

import freezegun
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from pycadf import cadftaxonomy

from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone import oauth1
from keystone.oauth1.backends import base
from keystone.tests import unit
from keystone.tests.unit.common import test_notifications
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile
from keystone.tests.unit import test_v3

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


def _urllib_parse_qs_text_keys(content):
    results = urllib.parse.parse_qs(content)
    return {key.decode('utf-8'): value for key, value in results.items()}


class OAuth1Tests(test_v3.RestfulTestCase):
    CONSUMER_URL = '/OS-OAUTH1/consumers'

    def setUp(self):
        super().setUp()
        # Now that the app has been served, we can query CONF values
        self.base_url = 'http://localhost/v3'

    def _create_single_consumer(self):
        ref = {'description': uuid.uuid4().hex}
        resp = self.post(self.CONSUMER_URL, body={'consumer': ref})
        return resp.result['consumer']

    def _create_request_token(self, consumer, project_id, base_url=None):
        endpoint = '/OS-OAUTH1/request_token'
        client = oauth1.Client(
            consumer['key'],
            client_secret=consumer['secret'],
            signature_method=oauth1.SIG_HMAC,
            callback_uri="oob",
        )
        headers = {'requested_project_id': project_id}
        if not base_url:
            base_url = self.base_url
        url, headers, body = client.sign(
            base_url + endpoint, http_method='POST', headers=headers
        )
        return endpoint, headers

    def _create_access_token(self, consumer, token, base_url=None):
        endpoint = '/OS-OAUTH1/access_token'
        client = oauth1.Client(
            consumer['key'],
            client_secret=consumer['secret'],
            resource_owner_key=token.key,
            resource_owner_secret=token.secret,
            signature_method=oauth1.SIG_HMAC,
            verifier=token.verifier,
        )
        if not base_url:
            base_url = self.base_url
        url, headers, body = client.sign(
            base_url + endpoint, http_method='POST'
        )
        headers.update({'Content-Type': 'application/json'})
        return endpoint, headers

    def _get_oauth_token(self, consumer, token):
        client = oauth1.Client(
            consumer['key'],
            client_secret=consumer['secret'],
            resource_owner_key=token.key,
            resource_owner_secret=token.secret,
            signature_method=oauth1.SIG_HMAC,
        )
        endpoint = '/auth/tokens'
        url, headers, body = client.sign(
            self.base_url + endpoint, http_method='POST'
        )
        headers.update({'Content-Type': 'application/json'})
        ref = {'auth': {'identity': {'oauth1': {}, 'methods': ['oauth1']}}}
        return endpoint, headers, ref

    def _authorize_request_token(self, request_id):
        if isinstance(request_id, bytes):
            request_id = request_id.decode()
        return f'/OS-OAUTH1/authorize/{request_id}'


class ConsumerCRUDTests(OAuth1Tests):
    def _consumer_create(
        self, description=None, description_flag=True, **kwargs
    ):
        if description_flag:
            ref = {'description': description}
        else:
            ref = {}
        if kwargs:
            ref.update(kwargs)
        resp = self.post(self.CONSUMER_URL, body={'consumer': ref})
        consumer = resp.result['consumer']
        consumer_id = consumer['id']
        self.assertEqual(description, consumer['description'])
        self.assertIsNotNone(consumer_id)
        self.assertIsNotNone(consumer['secret'])
        return consumer

    def test_consumer_create(self):
        description = uuid.uuid4().hex
        self._consumer_create(description=description)

    def test_consumer_create_none_desc_1(self):
        self._consumer_create()

    def test_consumer_create_none_desc_2(self):
        self._consumer_create(description_flag=False)

    def test_consumer_create_normalize_field(self):
        # If create a consumer with a field with : or - in the name,
        # the name is normalized by converting those chars to _.
        field_name = 'some:weird-field'
        field_value = uuid.uuid4().hex
        extra_fields = {field_name: field_value}
        consumer = self._consumer_create(**extra_fields)
        normalized_field_name = 'some_weird_field'
        self.assertEqual(field_value, consumer[normalized_field_name])

    def test_consumer_delete(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        resp = self.delete(self.CONSUMER_URL + f'/{consumer_id}')
        self.assertResponseStatus(resp, http.client.NO_CONTENT)

    def test_consumer_get_head(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        url = self.CONSUMER_URL + f'/{consumer_id}'
        resp = self.get(url)
        self_url = ['http://localhost/v3', self.CONSUMER_URL, '/', consumer_id]
        self_url = ''.join(self_url)
        self.assertEqual(self_url, resp.result['consumer']['links']['self'])
        self.assertEqual(consumer_id, resp.result['consumer']['id'])

        self.head(url, expected_status=http.client.OK)

    def test_consumer_list(self):
        self._consumer_create()
        resp = self.get(self.CONSUMER_URL)
        entities = resp.result['consumers']
        self.assertIsNotNone(entities)
        self_url = ['http://localhost/v3', self.CONSUMER_URL]
        self_url = ''.join(self_url)
        self.assertEqual(self_url, resp.result['links']['self'])
        self.assertValidListLinks(resp.result['links'])

        self.head(self.CONSUMER_URL, expected_status=http.client.OK)

    def test_consumer_update(self):
        consumer = self._create_single_consumer()
        original_id = consumer['id']
        original_description = consumer['description']
        update_description = original_description + '_new'

        update_ref = {'description': update_description}
        update_resp = self.patch(
            self.CONSUMER_URL + f'/{original_id}',
            body={'consumer': update_ref},
        )
        consumer = update_resp.result['consumer']
        self.assertEqual(update_description, consumer['description'])
        self.assertEqual(original_id, consumer['id'])

    def test_consumer_update_bad_secret(self):
        consumer = self._create_single_consumer()
        original_id = consumer['id']
        update_ref = copy.deepcopy(consumer)
        update_ref['description'] = uuid.uuid4().hex
        update_ref['secret'] = uuid.uuid4().hex
        self.patch(
            self.CONSUMER_URL + f'/{original_id}',
            body={'consumer': update_ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_consumer_update_bad_id(self):
        consumer = self._create_single_consumer()
        original_id = consumer['id']
        original_description = consumer['description']
        update_description = original_description + "_new"

        update_ref = copy.deepcopy(consumer)
        update_ref['description'] = update_description
        update_ref['id'] = update_description
        self.patch(
            self.CONSUMER_URL + f'/{original_id}',
            body={'consumer': update_ref},
            expected_status=http.client.BAD_REQUEST,
        )

    def test_consumer_update_normalize_field(self):
        # If update a consumer with a field with : or - in the name,
        # the name is normalized by converting those chars to _.
        field1_name = 'some:weird-field'
        field1_orig_value = uuid.uuid4().hex

        extra_fields = {field1_name: field1_orig_value}
        consumer = self._consumer_create(**extra_fields)
        consumer_id = consumer['id']

        field1_new_value = uuid.uuid4().hex

        field2_name = 'weird:some-field'
        field2_value = uuid.uuid4().hex

        update_ref = {field1_name: field1_new_value, field2_name: field2_value}

        update_resp = self.patch(
            self.CONSUMER_URL + f'/{consumer_id}',
            body={'consumer': update_ref},
        )
        consumer = update_resp.result['consumer']

        normalized_field1_name = 'some_weird_field'
        self.assertEqual(field1_new_value, consumer[normalized_field1_name])

        normalized_field2_name = 'weird_some_field'
        self.assertEqual(field2_value, consumer[normalized_field2_name])

    def test_consumer_create_no_description(self):
        resp = self.post(self.CONSUMER_URL, body={'consumer': {}})
        consumer = resp.result['consumer']
        consumer_id = consumer['id']
        self.assertIsNone(consumer['description'])
        self.assertIsNotNone(consumer_id)
        self.assertIsNotNone(consumer['secret'])

    def test_consumer_get_bad_id(self):
        url = self.CONSUMER_URL + f'/{uuid.uuid4().hex}'
        self.get(url, expected_status=http.client.NOT_FOUND)
        self.head(url, expected_status=http.client.NOT_FOUND)


class OAuthFlowTests(OAuth1Tests):
    def test_oauth_flow(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        self.consumer = {'key': consumer_id, 'secret': consumer_secret}
        self.assertIsNotNone(self.consumer['secret'])

        url, headers = self._create_request_token(
            self.consumer, self.project_id
        )
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = _urllib_parse_qs_text_keys(content.result)
        request_key = credentials['oauth_token'][0]
        request_secret = credentials['oauth_token_secret'][0]
        self.request_token = oauth1.Token(request_key, request_secret)
        self.assertIsNotNone(self.request_token.key)

        url = self._authorize_request_token(request_key)
        body = {'roles': [{'id': self.role_id}]}
        resp = self.put(url, body=body, expected_status=http.client.OK)
        self.verifier = resp.result['token']['oauth_verifier']
        self.assertTrue(all(i in base.VERIFIER_CHARS for i in self.verifier))
        self.assertEqual(8, len(self.verifier))

        self.request_token.set_verifier(self.verifier)
        url, headers = self._create_access_token(
            self.consumer, self.request_token
        )
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = _urllib_parse_qs_text_keys(content.result)
        access_key = credentials['oauth_token'][0]
        access_secret = credentials['oauth_token_secret'][0]
        self.access_token = oauth1.Token(access_key, access_secret)
        self.assertIsNotNone(self.access_token.key)

        url, headers, body = self._get_oauth_token(
            self.consumer, self.access_token
        )
        content = self.post(url, headers=headers, body=body)
        self.keystone_token_id = content.headers['X-Subject-Token']
        self.keystone_token = content.result['token']
        self.assertIsNotNone(self.keystone_token_id)

        # add a new role assignment to ensure it is ignored in the access token
        new_role = {'id': uuid.uuid4().hex, 'name': uuid.uuid4().hex}
        PROVIDERS.role_api.create_role(new_role['id'], new_role)
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            user_id=self.user_id,
            project_id=self.project_id,
            role_id=new_role['id'],
        )
        content = self.post(url, headers=headers, body=body)
        token = content.result['token']
        token_roles = [r['id'] for r in token['roles']]
        self.assertIn(self.role_id, token_roles)
        self.assertNotIn(new_role['id'], token_roles)


class AccessTokenCRUDTests(OAuthFlowTests):
    def test_delete_access_token_dne(self):
        self.delete(
            f'/users/{self.user_id}/OS-OAUTH1/access_tokens/{uuid.uuid4().hex}',
            expected_status=http.client.NOT_FOUND,
        )

    def test_list_no_access_tokens(self):
        url = f'/users/{self.user_id}/OS-OAUTH1/access_tokens'
        resp = self.get(url)
        entities = resp.result['access_tokens']
        self.assertEqual([], entities)
        self.assertValidListLinks(resp.result['links'])

        self.head(url, expected_status=http.client.OK)

    def test_get_single_access_token(self):
        self.test_oauth_flow()
        access_token_key_string = self.access_token.key.decode()

        url = f'/users/{self.user_id}/OS-OAUTH1/access_tokens/{access_token_key_string}'
        resp = self.get(url)
        entity = resp.result['access_token']
        self.assertEqual(access_token_key_string, entity['id'])
        self.assertEqual(self.consumer['key'], entity['consumer_id'])
        self.assertEqual('http://localhost/v3' + url, entity['links']['self'])

        self.head(url, expected_status=http.client.OK)

    def test_get_access_token_dne(self):
        url = (
            f'/users/{self.user_id}/OS-OAUTH1/access_tokens/{uuid.uuid4().hex}'
        )
        self.get(url, expected_status=http.client.NOT_FOUND)
        self.head(url, expected_status=http.client.NOT_FOUND)

    def test_list_all_roles_in_access_token(self):
        self.test_oauth_flow()
        url = f'/users/{self.user_id}/OS-OAUTH1/access_tokens/{self.access_token.key.decode()}/roles'
        resp = self.get(url)
        entities = resp.result['roles']
        self.assertTrue(entities)
        self.assertValidListLinks(resp.result['links'])

        self.head(url, expected_status=http.client.OK)

    def test_get_role_in_access_token(self):
        self.test_oauth_flow()

        access_token_key = self.access_token.key.decode()
        url = f'/users/{self.user_id}/OS-OAUTH1/access_tokens/{access_token_key}/roles/{self.role_id}'
        resp = self.get(url)
        entity = resp.result['role']
        self.assertEqual(self.role_id, entity['id'])

        self.head(url, expected_status=http.client.OK)

    def test_get_role_in_access_token_dne(self):
        self.test_oauth_flow()

        access_token_key = self.access_token.key.decode()
        url = f'/users/{self.user_id}/OS-OAUTH1/access_tokens/{access_token_key}/roles/{uuid.uuid4().hex}'
        self.get(url, expected_status=http.client.NOT_FOUND)
        self.head(url, expected_status=http.client.NOT_FOUND)

    def test_list_and_delete_access_tokens(self):
        self.test_oauth_flow()
        # List access_tokens should be > 0
        url = f'/users/{self.user_id}/OS-OAUTH1/access_tokens'
        resp = self.get(url)
        self.head(url, expected_status=http.client.OK)
        entities = resp.result['access_tokens']
        self.assertTrue(entities)
        self.assertValidListLinks(resp.result['links'])

        access_token_key = self.access_token.key.decode()
        # Delete access_token
        resp = self.delete(
            f'/users/{self.user_id}/OS-OAUTH1/access_tokens/{access_token_key}'
        )
        self.assertResponseStatus(resp, http.client.NO_CONTENT)

        # List access_token should be 0
        resp = self.get(url)
        self.head(url, expected_status=http.client.OK)
        entities = resp.result['access_tokens']
        self.assertEqual([], entities)
        self.assertValidListLinks(resp.result['links'])


class AuthTokenTests:
    def test_keystone_token_is_valid(self):
        self.test_oauth_flow()
        headers = {
            'X-Subject-Token': self.keystone_token_id,
            'X-Auth-Token': self.keystone_token_id,
        }
        r = self.get('/auth/tokens', headers=headers)
        self.assertValidTokenResponse(r, self.user)

        # now verify the oauth section
        oauth_section = r.result['token']['OS-OAUTH1']
        self.assertEqual(
            self.access_token.key.decode(), oauth_section['access_token_id']
        )
        self.assertEqual(self.consumer['key'], oauth_section['consumer_id'])

        # verify the roles section
        roles_list = r.result['token']['roles']
        # we can just verify the 0th role since we are only assigning one role
        self.assertEqual(self.role_id, roles_list[0]['id'])

        # verify that the token can perform delegated tasks
        ref = unit.new_user_ref(domain_id=self.domain_id)
        r = self.admin_request(
            path='/v3/users',
            headers=headers,
            method='POST',
            body={'user': ref},
        )
        self.assertValidUserResponse(r, ref)

    def test_delete_access_token_also_revokes_token(self):
        self.test_oauth_flow()

        access_token_key = self.access_token.key.decode()
        # Delete access token
        resp = self.delete(
            f'/users/{self.user_id}/OS-OAUTH1/access_tokens/{access_token_key}'
        )
        self.assertResponseStatus(resp, http.client.NO_CONTENT)

        # Check Keystone Token no longer exists
        headers = {
            'X-Subject-Token': self.keystone_token_id,
            'X-Auth-Token': self.keystone_token_id,
        }
        self.get(
            '/auth/tokens',
            headers=headers,
            expected_status=http.client.NOT_FOUND,
        )

    def test_deleting_consumer_also_deletes_tokens(self):
        self.test_oauth_flow()

        # Delete consumer
        consumer_id = self.consumer['key']
        resp = self.delete(f'/OS-OAUTH1/consumers/{consumer_id}')
        self.assertResponseStatus(resp, http.client.NO_CONTENT)

        # List access_token should be 0
        resp = self.get(f'/users/{self.user_id}/OS-OAUTH1/access_tokens')
        entities = resp.result['access_tokens']
        self.assertEqual([], entities)

        # Check Keystone Token no longer exists
        headers = {
            'X-Subject-Token': self.keystone_token_id,
            'X-Auth-Token': self.keystone_token_id,
        }
        self.head(
            '/auth/tokens',
            headers=headers,
            expected_status=http.client.NOT_FOUND,
        )

    def test_change_user_password_also_deletes_tokens(self):
        self.test_oauth_flow()

        # delegated keystone token exists
        headers = {
            'X-Subject-Token': self.keystone_token_id,
            'X-Auth-Token': self.keystone_token_id,
        }
        r = self.get('/auth/tokens', headers=headers)
        self.assertValidTokenResponse(r, self.user)

        user = {'password': uuid.uuid4().hex}
        r = self.patch(
            '/users/{user_id}'.format(user_id=self.user['id']),
            body={'user': user},
        )
        headers = {'X-Subject-Token': self.keystone_token_id}
        self.get(
            path='/auth/tokens',
            token=self.get_admin_token(),
            headers=headers,
            expected_status=http.client.NOT_FOUND,
        )

    def test_deleting_project_also_invalidates_tokens(self):
        self.test_oauth_flow()

        # delegated keystone token exists
        headers = {
            'X-Subject-Token': self.keystone_token_id,
            'X-Auth-Token': self.keystone_token_id,
        }
        r = self.get('/auth/tokens', headers=headers)
        self.assertValidTokenResponse(r, self.user)

        r = self.delete(f'/projects/{self.project_id}')

        headers = {'X-Subject-Token': self.keystone_token_id}
        self.get(
            path='/auth/tokens',
            token=self.get_admin_token(),
            headers=headers,
            expected_status=http.client.NOT_FOUND,
        )

    def test_token_chaining_is_not_allowed(self):
        self.test_oauth_flow()

        # attempt to re-authenticate (token chain) with the given token
        path = '/v3/auth/tokens/'
        auth_data = self.build_authentication_request(
            token=self.keystone_token_id
        )

        self.admin_request(
            path=path,
            body=auth_data,
            token=self.keystone_token_id,
            method='POST',
            expected_status=http.client.FORBIDDEN,
        )

    def test_delete_keystone_tokens_by_consumer_id(self):
        self.test_oauth_flow()
        PROVIDERS.token_provider_api._persistence.get_token(
            self.keystone_token_id
        )
        PROVIDERS.token_provider_api._persistence.delete_tokens(
            self.user_id, consumer_id=self.consumer['key']
        )
        self.assertRaises(
            exception.TokenNotFound,
            PROVIDERS.token_provider_api._persistence.get_token,
            self.keystone_token_id,
        )

    def _create_trust_get_token(self):
        ref = unit.new_trust_ref(
            trustor_user_id=self.user_id,
            trustee_user_id=self.user_id,
            project_id=self.project_id,
            impersonation=True,
            expires={'minutes': 1},
            role_ids=[self.role_id],
        )
        del ref['id']

        r = self.post('/OS-TRUST/trusts', body={'trust': ref})
        trust = self.assertValidTrustResponse(r)

        auth_data = self.build_authentication_request(
            user_id=self.user['id'],
            password=self.user['password'],
            trust_id=trust['id'],
        )

        return self.get_requested_token(auth_data)

    def _approve_request_token_url(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        self.consumer = {'key': consumer_id, 'secret': consumer_secret}
        self.assertIsNotNone(self.consumer['secret'])

        url, headers = self._create_request_token(
            self.consumer, self.project_id
        )
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = _urllib_parse_qs_text_keys(content.result)
        request_key = credentials['oauth_token'][0]
        request_secret = credentials['oauth_token_secret'][0]
        self.request_token = oauth1.Token(request_key, request_secret)
        self.assertIsNotNone(self.request_token.key)

        url = self._authorize_request_token(request_key)

        return url

    def test_oauth_token_cannot_create_new_trust(self):
        self.test_oauth_flow()
        ref = unit.new_trust_ref(
            trustor_user_id=self.user_id,
            trustee_user_id=self.user_id,
            project_id=self.project_id,
            impersonation=True,
            expires={'minutes': 1},
            role_ids=[self.role_id],
        )
        del ref['id']

        self.post(
            '/OS-TRUST/trusts',
            body={'trust': ref},
            token=self.keystone_token_id,
            expected_status=http.client.FORBIDDEN,
        )

    def test_oauth_token_cannot_authorize_request_token(self):
        self.test_oauth_flow()
        url = self._approve_request_token_url()
        body = {'roles': [{'id': self.role_id}]}
        self.put(
            url,
            body=body,
            token=self.keystone_token_id,
            expected_status=http.client.FORBIDDEN,
        )

    def test_oauth_token_cannot_list_request_tokens(self):
        self._set_policy(
            {
                "identity:list_access_tokens": [],
                "identity:create_consumer": [],
                "identity:authorize_request_token": [],
            }
        )
        self.test_oauth_flow()
        url = f'/users/{self.user_id}/OS-OAUTH1/access_tokens'
        self.get(
            url,
            token=self.keystone_token_id,
            expected_status=http.client.FORBIDDEN,
        )

    def _set_policy(self, new_policy):
        self.tempfile = self.useFixture(temporaryfile.SecureTempFile())
        self.tmpfilename = self.tempfile.file_name
        self.config_fixture.config(
            group='oslo_policy', policy_file=self.tmpfilename
        )
        with open(self.tmpfilename, "w") as policyfile:
            policyfile.write(jsonutils.dumps(new_policy))

    def test_trust_token_cannot_authorize_request_token(self):
        trust_token = self._create_trust_get_token()
        url = self._approve_request_token_url()
        body = {'roles': [{'id': self.role_id}]}
        self.put(
            url,
            body=body,
            token=trust_token,
            expected_status=http.client.FORBIDDEN,
        )

    def test_trust_token_cannot_list_request_tokens(self):
        self._set_policy(
            {"identity:list_access_tokens": [], "identity:create_trust": []}
        )
        trust_token = self._create_trust_get_token()
        url = f'/users/{self.user_id}/OS-OAUTH1/access_tokens'
        self.get(url, token=trust_token, expected_status=http.client.FORBIDDEN)


class FernetAuthTokenTests(AuthTokenTests, OAuthFlowTests):
    def config_overrides(self):
        super().config_overrides()
        self.config_fixture.config(group='token', provider='fernet')
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'fernet_tokens',
                CONF.fernet_tokens.max_active_keys,
            )
        )

    def test_delete_keystone_tokens_by_consumer_id(self):
        self.skipTest('Fernet tokens are never persisted in the backend.')


class MaliciousOAuth1Tests(OAuth1Tests):
    def _switch_baseurl_scheme(self):
        """Switch the base url scheme."""
        base_url_list = list(urlparse.urlparse(self.base_url))
        base_url_list[0] = 'https' if base_url_list[0] == 'http' else 'http'
        bad_url = urlparse.urlunparse(base_url_list)
        return bad_url

    def test_bad_consumer_secret(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer = {'key': consumer_id, 'secret': uuid.uuid4().hex}
        url, headers = self._create_request_token(consumer, self.project_id)
        self.post(
            url, headers=headers, expected_status=http.client.UNAUTHORIZED
        )

    def test_bad_request_url(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}
        bad_base_url = 'http://localhost/identity_admin/v3'
        url, headers = self._create_request_token(
            consumer, self.project_id, base_url=bad_base_url
        )
        self.post(
            url, headers=headers, expected_status=http.client.UNAUTHORIZED
        )

    def test_bad_request_url_scheme(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}
        bad_url_scheme = self._switch_baseurl_scheme()
        url, headers = self._create_request_token(
            consumer, self.project_id, base_url=bad_url_scheme
        )
        self.post(
            url, headers=headers, expected_status=http.client.UNAUTHORIZED
        )

    def test_bad_request_token_key(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}
        url, headers = self._create_request_token(consumer, self.project_id)
        self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        url = self._authorize_request_token(uuid.uuid4().hex)
        body = {'roles': [{'id': self.role_id}]}
        self.put(url, body=body, expected_status=http.client.NOT_FOUND)

    def test_bad_request_body_when_authorize(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}
        url, headers = self._create_request_token(consumer, self.project_id)
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = _urllib_parse_qs_text_keys(content.result)
        request_key = credentials['oauth_token'][0]
        url = self._authorize_request_token(request_key)
        bad_body = {'roles': [{'fake_key': 'fake_value'}]}
        self.put(url, body=bad_body, expected_status=http.client.BAD_REQUEST)

    def test_bad_consumer_id(self):
        consumer = self._create_single_consumer()
        consumer_id = uuid.uuid4().hex
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}
        url, headers = self._create_request_token(consumer, self.project_id)
        self.post(url, headers=headers, expected_status=http.client.NOT_FOUND)

    def test_bad_requested_project_id(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}
        project_id = uuid.uuid4().hex
        url, headers = self._create_request_token(consumer, project_id)
        self.post(url, headers=headers, expected_status=http.client.NOT_FOUND)

    def test_bad_verifier(self):
        self.config_fixture.config(debug=True, insecure_debug=True)
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}

        url, headers = self._create_request_token(consumer, self.project_id)
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = _urllib_parse_qs_text_keys(content.result)
        request_key = credentials['oauth_token'][0]
        request_secret = credentials['oauth_token_secret'][0]
        request_token = oauth1.Token(request_key, request_secret)

        url = self._authorize_request_token(request_key)
        body = {'roles': [{'id': self.role_id}]}
        resp = self.put(url, body=body, expected_status=http.client.OK)
        verifier = resp.result['token']['oauth_verifier']
        self.assertIsNotNone(verifier)

        request_token.set_verifier(uuid.uuid4().hex)
        url, headers = self._create_access_token(consumer, request_token)
        resp = self.post(
            url, headers=headers, expected_status=http.client.BAD_REQUEST
        )
        resp_data = jsonutils.loads(resp.body)
        self.assertIn(
            'Validation failed with errors',
            resp_data.get('error', {}).get('message'),
        )

    def test_validate_access_token_request_failed(self):
        self.config_fixture.config(debug=True, insecure_debug=True)
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}

        url, headers = self._create_request_token(consumer, self.project_id)
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = _urllib_parse_qs_text_keys(content.result)
        request_key = credentials['oauth_token'][0]
        request_secret = credentials['oauth_token_secret'][0]
        request_token = oauth1.Token(request_key, request_secret)

        url = self._authorize_request_token(request_key)
        body = {'roles': [{'id': self.role_id}]}
        resp = self.put(url, body=body, expected_status=http.client.OK)
        verifier = resp.result['token']['oauth_verifier']
        request_token.set_verifier(verifier)

        # 1. Invalid base url.
        # Update the base url, so it will fail to validate the signature.
        base_url = 'http://localhost/identity_admin/v3'
        url, headers = self._create_access_token(
            consumer, request_token, base_url=base_url
        )
        resp = self.post(
            url, headers=headers, expected_status=http.client.UNAUTHORIZED
        )
        resp_data = jsonutils.loads(resp.body)
        self.assertIn(
            'Invalid signature', resp_data.get('error', {}).get('message')
        )

        # 2. Invalid base url scheme.
        # Update the base url scheme, so it will fail to validate signature.
        bad_url_scheme = self._switch_baseurl_scheme()
        url, headers = self._create_access_token(
            consumer, request_token, base_url=bad_url_scheme
        )
        resp = self.post(
            url, headers=headers, expected_status=http.client.UNAUTHORIZED
        )
        resp_data = jsonutils.loads(resp.body)
        self.assertIn(
            'Invalid signature', resp_data.get('error', {}).get('message')
        )

        # 3. Invalid signature.
        # Update the secret, so it will fail to validate the signature.
        consumer.update({'secret': uuid.uuid4().hex})
        url, headers = self._create_access_token(consumer, request_token)
        resp = self.post(
            url, headers=headers, expected_status=http.client.UNAUTHORIZED
        )
        resp_data = jsonutils.loads(resp.body)
        self.assertIn(
            'Invalid signature', resp_data.get('error', {}).get('message')
        )

        # 4. Invalid verifier.
        # Even though the verifier is well formatted, it is not verifier
        # that is stored in the backend, this is different with the testcase
        # above `test_bad_verifier` where it test that `verifier` is not
        # well formatted.
        verifier = ''.join(
            random.SystemRandom().sample(base.VERIFIER_CHARS, 8)
        )
        request_token.set_verifier(verifier)
        url, headers = self._create_access_token(consumer, request_token)
        resp = self.post(
            url, headers=headers, expected_status=http.client.UNAUTHORIZED
        )
        resp_data = jsonutils.loads(resp.body)
        self.assertIn(
            'Provided verifier', resp_data.get('error', {}).get('message')
        )

        # 5. The provided consumer does not exist.
        consumer.update({'key': uuid.uuid4().hex})
        url, headers = self._create_access_token(consumer, request_token)
        resp = self.post(
            url, headers=headers, expected_status=http.client.UNAUTHORIZED
        )
        resp_data = jsonutils.loads(resp.body)
        self.assertIn(
            'Provided consumer does not exist',
            resp_data.get('error', {}).get('message'),
        )

        # 6. The consumer key provided does not match stored consumer key.
        consumer2 = self._create_single_consumer()
        consumer.update({'key': consumer2['id']})
        url, headers = self._create_access_token(consumer, request_token)
        resp = self.post(
            url, headers=headers, expected_status=http.client.UNAUTHORIZED
        )
        resp_data = jsonutils.loads(resp.body)
        self.assertIn(
            'Provided consumer key', resp_data.get('error', {}).get('message')
        )

    def test_bad_authorizing_roles_id(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}

        # This new role is utilzied to ensure the user still has access to
        # the project but is authorizing an incorrect role_id for the purposes
        # of oauth1.
        new_role = {'id': uuid.uuid4().hex, 'name': uuid.uuid4().hex}
        PROVIDERS.role_api.create_role(new_role['id'], new_role)
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            user_id=self.user_id,
            project_id=self.project_id,
            role_id=new_role['id'],
        )

        url, headers = self._create_request_token(consumer, self.project_id)
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = _urllib_parse_qs_text_keys(content.result)
        request_key = credentials['oauth_token'][0]

        PROVIDERS.assignment_api.remove_role_from_user_and_project(
            self.user_id, self.project_id, new_role['id']
        )
        url = self._authorize_request_token(request_key)
        body = {'roles': [{'id': new_role['id']}]}
        # NOTE(morgan): previous versions of this test erroneously checked for
        # 404 because an unrouted URI was being hit. It is correct to get a 401
        # error back as the role is not in the superset of roles the user
        # has at the time of the Authorization.
        self.put(path=url, body=body, expected_status=http.client.UNAUTHORIZED)

    def test_bad_authorizing_roles_name(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}

        url, headers = self._create_request_token(consumer, self.project_id)
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = _urllib_parse_qs_text_keys(content.result)
        request_key = credentials['oauth_token'][0]

        url = self._authorize_request_token(request_key)
        body = {'roles': [{'name': 'fake_name'}]}
        self.put(path=url, body=body, expected_status=http.client.NOT_FOUND)

    def test_no_authorizing_user_id(self):
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}

        url, headers = self._create_request_token(consumer, self.project_id)
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = _urllib_parse_qs_text_keys(content.result)
        request_key = credentials['oauth_token'][0]
        request_secret = credentials['oauth_token_secret'][0]
        request_token = oauth1.Token(request_key, request_secret)

        url = self._authorize_request_token(request_key)
        body = {'roles': [{'id': self.role_id}]}
        resp = self.put(url, body=body, expected_status=http.client.OK)
        verifier = resp.result['token']['oauth_verifier']
        request_token.set_verifier(verifier)
        request_token_created = PROVIDERS.oauth_api.get_request_token(
            request_key.decode('utf-8')
        )
        request_token_created.update({'authorizing_user_id': ''})
        # Update the request token that is created instead of mocking
        # the whole token object to focus on what's we want to test
        # here and avoid any other factors that will result in the same
        # exception.
        with mock.patch.object(
            PROVIDERS.oauth_api, 'get_request_token'
        ) as mock_token:
            mock_token.return_value = request_token_created
            url, headers = self._create_access_token(consumer, request_token)
            self.post(
                url, headers=headers, expected_status=http.client.UNAUTHORIZED
            )

    def test_validate_requet_token_request_failed(self):
        self.config_fixture.config(debug=True, insecure_debug=True)
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        consumer = {'key': consumer_id, 'secret': consumer_secret}

        url = '/OS-OAUTH1/request_token'
        auth_header = (
            'OAuth oauth_version="1.0", oauth_consumer_key=' + consumer_id
        )
        faked_header = {
            'Authorization': auth_header,
            'requested_project_id': self.project_id,
        }

        resp = self.post(
            url, headers=faked_header, expected_status=http.client.BAD_REQUEST
        )
        resp_data = jsonutils.loads(resp.body)
        self.assertIn(
            'Validation failed with errors', resp_data['error']['message']
        )

    def test_expired_authorizing_request_token(self):
        with freezegun.freeze_time(timeutils.utcnow()) as frozen_time:
            self.config_fixture.config(
                group='oauth1', request_token_duration=1
            )

            consumer = self._create_single_consumer()
            consumer_id = consumer['id']
            consumer_secret = consumer['secret']
            self.consumer = {'key': consumer_id, 'secret': consumer_secret}
            self.assertIsNotNone(self.consumer['key'])

            url, headers = self._create_request_token(
                self.consumer, self.project_id
            )
            content = self.post(
                url,
                headers=headers,
                response_content_type='application/x-www-form-urlencoded',
            )
            credentials = _urllib_parse_qs_text_keys(content.result)
            request_key = credentials['oauth_token'][0]
            request_secret = credentials['oauth_token_secret'][0]
            self.request_token = oauth1.Token(request_key, request_secret)
            self.assertIsNotNone(self.request_token.key)

            url = self._authorize_request_token(request_key)
            body = {'roles': [{'id': self.role_id}]}
            frozen_time.tick(
                delta=datetime.timedelta(
                    seconds=CONF.oauth1.request_token_duration + 1
                )
            )
            self.put(url, body=body, expected_status=http.client.UNAUTHORIZED)

    def test_expired_creating_keystone_token(self):
        with freezegun.freeze_time(timeutils.utcnow()) as frozen_time:
            self.config_fixture.config(group='oauth1', access_token_duration=1)

            consumer = self._create_single_consumer()
            consumer_id = consumer['id']
            consumer_secret = consumer['secret']
            self.consumer = {'key': consumer_id, 'secret': consumer_secret}
            self.assertIsNotNone(self.consumer['key'])

            url, headers = self._create_request_token(
                self.consumer, self.project_id
            )
            content = self.post(
                url,
                headers=headers,
                response_content_type='application/x-www-form-urlencoded',
            )
            credentials = _urllib_parse_qs_text_keys(content.result)
            request_key = credentials['oauth_token'][0]
            request_secret = credentials['oauth_token_secret'][0]
            self.request_token = oauth1.Token(request_key, request_secret)
            self.assertIsNotNone(self.request_token.key)

            url = self._authorize_request_token(request_key)
            body = {'roles': [{'id': self.role_id}]}
            resp = self.put(url, body=body, expected_status=http.client.OK)
            self.verifier = resp.result['token']['oauth_verifier']

            self.request_token.set_verifier(self.verifier)
            url, headers = self._create_access_token(
                self.consumer, self.request_token
            )
            content = self.post(
                url,
                headers=headers,
                response_content_type='application/x-www-form-urlencoded',
            )
            credentials = _urllib_parse_qs_text_keys(content.result)
            access_key = credentials['oauth_token'][0]
            access_secret = credentials['oauth_token_secret'][0]
            self.access_token = oauth1.Token(access_key, access_secret)
            self.assertIsNotNone(self.access_token.key)

            url, headers, body = self._get_oauth_token(
                self.consumer, self.access_token
            )
            frozen_time.tick(
                delta=datetime.timedelta(
                    seconds=CONF.oauth1.access_token_duration + 1
                )
            )
            self.post(
                url,
                headers=headers,
                body=body,
                expected_status=http.client.UNAUTHORIZED,
            )

    def test_missing_oauth_headers(self):
        endpoint = '/OS-OAUTH1/request_token'
        client = oauth1.Client(
            uuid.uuid4().hex,
            client_secret=uuid.uuid4().hex,
            signature_method=oauth1.SIG_HMAC,
            callback_uri="oob",
        )
        headers = {'requested_project_id': uuid.uuid4().hex}
        _url, headers, _body = client.sign(
            self.base_url + endpoint, http_method='POST', headers=headers
        )

        # NOTE(stevemar): To simulate this error, we remove the Authorization
        # header from the post request.
        del headers['Authorization']
        self.post(
            endpoint,
            headers=headers,
            expected_status=http.client.INTERNAL_SERVER_ERROR,
        )


class OAuthNotificationTests(
    OAuth1Tests, test_notifications.BaseNotificationTest
):
    def test_create_consumer(self):
        consumer_ref = self._create_single_consumer()
        self._assert_notify_sent(
            consumer_ref['id'],
            test_notifications.CREATED_OPERATION,
            'OS-OAUTH1:consumer',
        )
        self._assert_last_audit(
            consumer_ref['id'],
            test_notifications.CREATED_OPERATION,
            'OS-OAUTH1:consumer',
            cadftaxonomy.SECURITY_ACCOUNT,
        )

    def test_update_consumer(self):
        consumer_ref = self._create_single_consumer()
        update_ref = {'consumer': {'description': uuid.uuid4().hex}}
        PROVIDERS.oauth_api.update_consumer(consumer_ref['id'], update_ref)
        self._assert_notify_sent(
            consumer_ref['id'],
            test_notifications.UPDATED_OPERATION,
            'OS-OAUTH1:consumer',
        )
        self._assert_last_audit(
            consumer_ref['id'],
            test_notifications.UPDATED_OPERATION,
            'OS-OAUTH1:consumer',
            cadftaxonomy.SECURITY_ACCOUNT,
        )

    def test_delete_consumer(self):
        consumer_ref = self._create_single_consumer()
        PROVIDERS.oauth_api.delete_consumer(consumer_ref['id'])
        self._assert_notify_sent(
            consumer_ref['id'],
            test_notifications.DELETED_OPERATION,
            'OS-OAUTH1:consumer',
        )
        self._assert_last_audit(
            consumer_ref['id'],
            test_notifications.DELETED_OPERATION,
            'OS-OAUTH1:consumer',
            cadftaxonomy.SECURITY_ACCOUNT,
        )

    def test_oauth_flow_notifications(self):
        """Test to ensure notifications are sent for oauth tokens.

        This test is very similar to test_oauth_flow, however
        there are additional checks in this test for ensuring that
        notifications for request token creation, and access token
        creation/deletion are emitted.
        """
        consumer = self._create_single_consumer()
        consumer_id = consumer['id']
        consumer_secret = consumer['secret']
        self.consumer = {'key': consumer_id, 'secret': consumer_secret}
        self.assertIsNotNone(self.consumer['secret'])

        url, headers = self._create_request_token(
            self.consumer, self.project_id
        )
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = _urllib_parse_qs_text_keys(content.result)
        request_key = credentials['oauth_token'][0]
        request_secret = credentials['oauth_token_secret'][0]
        self.request_token = oauth1.Token(request_key, request_secret)
        self.assertIsNotNone(self.request_token.key)

        request_key_string = request_key.decode()
        # Test to ensure the create request token notification is sent
        self._assert_notify_sent(
            request_key_string,
            test_notifications.CREATED_OPERATION,
            'OS-OAUTH1:request_token',
        )
        self._assert_last_audit(
            request_key_string,
            test_notifications.CREATED_OPERATION,
            'OS-OAUTH1:request_token',
            cadftaxonomy.SECURITY_CREDENTIAL,
        )

        url = self._authorize_request_token(request_key)
        body = {'roles': [{'id': self.role_id}]}
        resp = self.put(url, body=body, expected_status=http.client.OK)
        self.verifier = resp.result['token']['oauth_verifier']
        self.assertTrue(all(i in base.VERIFIER_CHARS for i in self.verifier))
        self.assertEqual(8, len(self.verifier))

        self.request_token.set_verifier(self.verifier)
        url, headers = self._create_access_token(
            self.consumer, self.request_token
        )
        content = self.post(
            url,
            headers=headers,
            response_content_type='application/x-www-form-urlencoded',
        )
        credentials = _urllib_parse_qs_text_keys(content.result)
        access_key = credentials['oauth_token'][0]
        access_secret = credentials['oauth_token_secret'][0]
        self.access_token = oauth1.Token(access_key, access_secret)
        self.assertIsNotNone(self.access_token.key)

        access_key_string = access_key.decode()
        # Test to ensure the create access token notification is sent
        self._assert_notify_sent(
            access_key_string,
            test_notifications.CREATED_OPERATION,
            'OS-OAUTH1:access_token',
        )
        self._assert_last_audit(
            access_key_string,
            test_notifications.CREATED_OPERATION,
            'OS-OAUTH1:access_token',
            cadftaxonomy.SECURITY_CREDENTIAL,
        )

        resp = self.delete(
            f'/users/{self.user_id}/OS-OAUTH1/access_tokens/{self.access_token.key.decode()}'
        )
        self.assertResponseStatus(resp, http.client.NO_CONTENT)

        # Test to ensure the delete access token notification is sent
        self._assert_notify_sent(
            access_key_string,
            test_notifications.DELETED_OPERATION,
            'OS-OAUTH1:access_token',
        )
        self._assert_last_audit(
            access_key_string,
            test_notifications.DELETED_OPERATION,
            'OS-OAUTH1:access_token',
            cadftaxonomy.SECURITY_CREDENTIAL,
        )


class OAuthCADFNotificationTests(OAuthNotificationTests):
    def setUp(self):
        """Repeat the tests for CADF notifications."""
        super().setUp()
        self.config_fixture.config(notification_format='cadf')


class JsonHomeTests(OAuth1Tests, test_v3.JsonHomeTestMixin):
    JSON_HOME_DATA = {
        'https://docs.openstack.org/api/openstack-identity/3/ext/OS-OAUTH1/1.0'
        '/rel/consumers': {'href': '/OS-OAUTH1/consumers'}
    }



# Copyright 2013 Metacloud
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Keystone Caching Layer Implementation."""

import secrets

from dogpile.cache import region
from dogpile.cache import util
from oslo_cache import core as cache

from keystone.common.cache import _context_cache
import keystone.conf

CONF = keystone.conf.CONF


class RegionInvalidationManager:
    REGION_KEY_PREFIX = '<<<region>>>:'

    def __init__(self, invalidation_region, region_name):
        self._invalidation_region = invalidation_region
        self._region_key = self.REGION_KEY_PREFIX + region_name

    def _generate_new_id(self):
        return secrets.token_bytes(10)

    @property
    def region_id(self):
        return self._invalidation_region.get_or_create(
            self._region_key, self._generate_new_id, expiration_time=-1
        )

    def invalidate_region(self):
        new_region_id = self._generate_new_id()
        self._invalidation_region.set(self._region_key, new_region_id)
        return new_region_id

    def is_region_key(self, key):
        return key == self._region_key


class DistributedInvalidationStrategy(region.RegionInvalidationStrategy):
    def __init__(self, region_manager):
        self._region_manager = region_manager

    def invalidate(self, hard=None):
        self._region_manager.invalidate_region()

    def is_invalidated(self, timestamp):
        return False

    def was_hard_invalidated(self):
        return False

    def is_hard_invalidated(self, timestamp):
        return False

    def was_soft_invalidated(self):
        return False

    def is_soft_invalidated(self, timestamp):
        return False


def key_mangler_factory(invalidation_manager, orig_key_mangler):
    def key_mangler(key):
        # NOTE(dstanek): Since *all* keys go through the key mangler we
        # need to make sure the region keys don't get the region_id added.
        # If it were there would be no way to get to it, making the cache
        # effectively useless.
        if not invalidation_manager.is_region_key(key):
            key = f'{key}:{invalidation_manager.region_id}'
        if orig_key_mangler:
            key = orig_key_mangler(key)
        return key

    return key_mangler


def create_region(name):
    """Create a dopile region.

    Wraps oslo_cache.core.create_region. This is used to ensure that the
    Region is properly patched and allows us to more easily specify a region
    name.

    :param str name: The region name
    :returns: The new region.
    :rtype: :class:`dogpile.cache.region.CacheRegion`

    """
    region = cache.create_region()
    region.name = name  # oslo.cache doesn't allow this yet
    return region


CACHE_REGION = create_region(name='shared default')
CACHE_INVALIDATION_REGION = create_region(name='invalidation region')

register_model_handler = _context_cache._register_model_handler


def configure_cache(region=None):
    if region is None:
        region = CACHE_REGION
    # NOTE(morganfainberg): running cache.configure_cache_region()
    # sets region.is_configured, this must be captured before
    # cache.configure_cache_region is called.
    configured = region.is_configured
    cache.configure_cache_region(CONF, region)
    # Only wrap the region if it was not configured. This should be pushed
    # to oslo_cache lib somehow.
    if not configured:
        region.wrap(_context_cache._ResponseCacheProxy)

        region_manager = RegionInvalidationManager(
            CACHE_INVALIDATION_REGION, region.name
        )
        region.key_mangler = key_mangler_factory(
            region_manager, region.key_mangler
        )
        region.region_invalidator = DistributedInvalidationStrategy(
            region_manager
        )


def _sha1_mangle_key(key):
    """Wrapper for dogpile's sha1_mangle_key.

    dogpile's sha1_mangle_key function expects an encoded string, so we
    should take steps to properly handle multiple inputs before passing
    the key through.

    NOTE(dstanek): this was copied directly from olso_cache
    """
    try:
        key = key.encode('utf-8', errors='xmlcharrefreplace')
    except (UnicodeError, AttributeError):
        # NOTE(stevemar): if encoding fails just continue anyway.
        pass
    return util.sha1_mangle_key(key)


def configure_invalidation_region():
    if CACHE_INVALIDATION_REGION.is_configured:
        return

    # NOTE(dstanek): Configuring this region manually so that we control the
    # expiration and can ensure that the keys don't expire.
    config_dict = cache._build_cache_config(CONF)
    config_dict['expiration_time'] = None  # we don't want an expiration

    CACHE_INVALIDATION_REGION.configure_from_config(
        config_dict, f'{CONF.cache.config_prefix}.'
    )

    # NOTE(breton): Wrap the cache invalidation region to avoid excessive
    # calls to memcached, which would result in poor performance.
    CACHE_INVALIDATION_REGION.wrap(_context_cache._ResponseCacheProxy)

    # NOTE(morganfainberg): if the backend requests the use of a
    # key_mangler, we should respect that key_mangler function.  If a
    # key_mangler is not defined by the backend, use the sha1_mangle_key
    # mangler provided by dogpile.cache. This ensures we always use a fixed
    # size cache-key.
    if CACHE_INVALIDATION_REGION.key_mangler is None:
        CACHE_INVALIDATION_REGION.key_mangler = _sha1_mangle_key


def get_memoization_decorator(group, expiration_group=None, region=None):
    if region is None:
        region = CACHE_REGION
    return cache.get_memoization_decorator(
        CONF, region, group, expiration_group=expiration_group
    )



# Copyright 2018 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Main entry point into the Application Credential service."""

from keystone.common import cache
from keystone.common import driver_hints
from keystone.common import manager
from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone import notifications

CONF = keystone.conf.CONF
MEMOIZE = cache.get_memoization_decorator(group='application_credential')
PROVIDERS = provider_api.ProviderAPIs


class Manager(manager.Manager):
    """Default pivot point for the Application Credential backend.

    See :mod:`keystone.common.manager.Manager` for more details on how this
    dynamically calls the backend.

    """

    driver_namespace = 'keystone.application_credential'
    _provides_api = 'application_credential_api'

    _APP_CRED = 'application_credential'
    _ACCESS_RULE = 'access_rule'

    def __init__(self):
        super().__init__(CONF.application_credential.driver)
        self._register_callback_listeners()

    def _register_callback_listeners(self):
        notifications.register_event_callback(
            notifications.ACTIONS.deleted,
            'user',
            self._delete_app_creds_on_user_delete_callback,
        )
        notifications.register_event_callback(
            notifications.ACTIONS.disabled,
            'user',
            self._delete_app_creds_on_user_delete_callback,
        )
        notifications.register_event_callback(
            notifications.ACTIONS.internal,
            notifications.REMOVE_APP_CREDS_FOR_USER,
            self._delete_app_creds_on_assignment_removal,
        )

    def _delete_app_creds_on_user_delete_callback(
        self, service, resource_type, operation, payload
    ):
        user_id = payload['resource_info']
        self._delete_application_credentials_for_user(user_id)
        self._delete_access_rules_for_user(user_id)

    def _delete_app_creds_on_assignment_removal(
        self, service, resource_type, operation, payload
    ):
        user_id = payload['resource_info']['user_id']
        project_id = payload['resource_info']['project_id']
        self._delete_application_credentials_for_user_on_project(
            user_id, project_id
        )

    def _get_user_roles(self, user_id, project_id):
        assignment_list = self.assignment_api.list_role_assignments(
            user_id=user_id, project_id=project_id, effective=True
        )
        return list({x['role_id'] for x in assignment_list})

    def _require_user_has_role_in_project(self, roles, user_id, project_id):
        user_roles = self._get_user_roles(user_id, project_id)
        for role in roles:
            if role['id'] not in user_roles:
                raise exception.RoleAssignmentNotFound(
                    role_id=role['id'], actor_id=user_id, target_id=project_id
                )

    def _assert_limit_not_exceeded(self, user_id):
        user_limit = CONF.application_credential.user_limit
        if user_limit >= 0:
            app_cred_count = len(self.list_application_credentials(user_id))
            if app_cred_count >= user_limit:
                raise exception.ApplicationCredentialLimitExceeded(
                    limit=user_limit
                )

    def _get_role_list(self, app_cred_roles):
        roles = []
        for role in app_cred_roles:
            roles.append(PROVIDERS.role_api.get_role(role['id']))
        return roles

    def authenticate(self, application_credential_id, secret):
        """Authenticate with an application credential.

        :param str application_credential_id: Application Credential ID
        :param str secret: Application Credential secret

        """
        self.driver.authenticate(application_credential_id, secret)

    def _process_app_cred(self, app_cred_ref):
        app_cred_ref = app_cred_ref.copy()
        app_cred_ref.pop('secret_hash')
        app_cred_ref['roles'] = self._get_role_list(app_cred_ref['roles'])
        return app_cred_ref

    def create_application_credential(
        self, application_credential, initiator=None
    ):
        """Create a new application credential.

        :param dict application_credential: Application Credential data
        :param initiator: CADF initiator

        :returns: a new application credential
        """
        application_credential = application_credential.copy()
        user_id = application_credential['user_id']
        project_id = application_credential['project_id']
        roles = application_credential.pop('roles', [])
        access_rules = application_credential.pop('access_rules', None)

        self._assert_limit_not_exceeded(user_id)
        self._require_user_has_role_in_project(roles, user_id, project_id)
        unhashed_secret = application_credential['secret']
        ref = self.driver.create_application_credential(
            application_credential, roles, access_rules
        )
        ref['secret'] = unhashed_secret
        ref = self._process_app_cred(ref)
        notifications.Audit.created(
            self._APP_CRED, application_credential['id'], initiator
        )
        return ref

    @MEMOIZE
    def get_application_credential(self, application_credential_id):
        """Get application credential details.

        :param str application_credential_id: Application Credential ID

        :returns: an application credential
        """
        app_cred = self.driver.get_application_credential(
            application_credential_id
        )
        return self._process_app_cred(app_cred)

    def list_application_credentials(self, user_id, hints=None):
        """List application credentials for a user.

        :param str user_id: User ID
        :param dict hints: Properties to filter on

        :returns: a list of application credentials
        """
        hints = hints or driver_hints.Hints()
        app_cred_list = self.driver.list_application_credentials_for_user(
            user_id, hints
        )
        return [self._process_app_cred(app_cred) for app_cred in app_cred_list]

    @MEMOIZE
    def get_access_rule(self, access_rule_id):
        """Get access rule details.

        :param str access_rule_id: Access Rule ID

        :returns: an access rule
        """
        return self.driver.get_access_rule(access_rule_id)

    def list_access_rules_for_user(self, user_id, hints=None):
        """List access rules for user.

        :param str user_id: User ID

        :returns: a list of access rules
        """
        hints = hints or driver_hints.Hints()
        return self.driver.list_access_rules_for_user(user_id, hints)

    def delete_application_credential(
        self, application_credential_id, initiator=None
    ):
        """Delete an application credential.

        :param str application_credential_id: Application Credential ID
        :param initiator: CADF initiator

        :raises keystone.exception.ApplicationCredentialNotFound: If the
            application credential doesn't exist.
        """
        self.driver.delete_application_credential(application_credential_id)
        self.get_application_credential.invalidate(
            self, application_credential_id
        )
        notifications.Audit.deleted(
            self._APP_CRED, application_credential_id, initiator
        )

    def _delete_application_credentials_for_user(
        self, user_id, initiator=None
    ):
        """Delete all application credentials for a user.

        :param str user_id: User ID

        This is triggered when a user is deleted.
        """
        app_creds = self.driver.list_application_credentials_for_user(
            user_id, driver_hints.Hints()
        )
        self.driver.delete_application_credentials_for_user(user_id)
        for app_cred in app_creds:
            self.get_application_credential.invalidate(self, app_cred['id'])
            notifications.Audit.deleted(
                self._APP_CRED, app_cred['id'], initiator
            )

    def _delete_application_credentials_for_user_on_project(
        self, user_id, project_id
    ):
        """Delete all application credentials for a user on a given project.

        :param str user_id: User ID
        :param str project_id: Project ID

        This is triggered when a user loses a role assignment on a project.
        """
        hints = driver_hints.Hints()
        hints.add_filter('project_id', project_id)
        app_creds = self.driver.list_application_credentials_for_user(
            user_id, hints
        )

        self.driver.delete_application_credentials_for_user_on_project(
            user_id, project_id
        )
        for app_cred in app_creds:
            self.get_application_credential.invalidate(self, app_cred['id'])

    def delete_access_rule(self, access_rule_id, initiator=None):
        """Delete an access rule.

        :param str: access_rule_id: Access Rule ID
        :param initiator: CADF initiator

        :raises keystone.exception.AccessRuleNotFound: If the access rule
            doesn't exist.
        """
        self.driver.delete_access_rule(access_rule_id)
        self.get_access_rule.invalidate(self, access_rule_id)
        notifications.Audit.deleted(
            self._ACCESS_RULE, access_rule_id, initiator
        )

    def _delete_access_rules_for_user(self, user_id, initiator=None):
        """Delete all access rules for a user.

        :param str user_id: User ID

        This is triggered when a user is deleted.
        """
        access_rules = self.driver.list_access_rules_for_user(
            user_id, driver_hints.Hints()
        )
        self.driver.delete_access_rules_for_user(user_id)
        for rule in access_rules:
            self.get_access_rule.invalidate(self, rule['id'])
            notifications.Audit.deleted(
                self._ACCESS_RULE, rule['id'], initiator
            )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This file handles all flask-restful resources for /OS-EP-FILTER

import http.client

import flask_restful

from keystone.api._shared import json_home_relations
from keystone.api import endpoints as _endpoints_api
from keystone.api import validation
from keystone.catalog import schema
from keystone.common import json_home
from keystone.common import provider_api
from keystone.common import rbac_enforcer
from keystone import exception
from keystone.i18n import _
from keystone.server import flask as ks_flask

ENFORCER = rbac_enforcer.RBACEnforcer
PROVIDERS = provider_api.ProviderAPIs

_build_resource_relation = json_home_relations.os_ep_filter_resource_rel_func
_build_parameter_relation = json_home_relations.os_ep_filter_parameter_rel_func

_ENDPOINT_GROUP_PARAMETER_RELATION = _build_parameter_relation(
    parameter_name='endpoint_group_id'
)


# NOTE(morgan): This is shared from keystone.api.endpoint, this is a special
# case where cross-api code is used. This pattern should not be replicated.
_filter_endpoint = _endpoints_api._filter_endpoint


def _require_valid_filter(endpoint_group):
    valid_filter_keys = ['service_id', 'region_id', 'interface']

    filters = endpoint_group.get('filters', {})
    for key in filters.keys():
        if key not in valid_filter_keys:
            raise exception.ValidationError(
                attribute=' or '.join(valid_filter_keys),
                target='endpoint_group',
            )


class EndpointGroupsResource(ks_flask.ResourceBase):
    collection_key = 'endpoint_groups'
    member_key = 'endpoint_group'
    api_prefix = '/OS-EP-FILTER'
    json_home_resource_rel_func = _build_resource_relation
    json_home_parameter_rel_func = _build_parameter_relation

    @validation.request_query_schema(schema.endpoint_group_index_request_query)
    @validation.response_body_schema(schema.endpoint_group_index_response_body)
    def get(self):
        """List all endpoint groups.

        GET /v3/OS-EP-FILTER/endpoint_groups
        """
        filters = 'name'
        ENFORCER.enforce_call(
            action='identity:list_endpoint_groups', filters=filters
        )
        hints = self.build_driver_hints(filters)
        refs = PROVIDERS.catalog_api.list_endpoint_groups(hints)
        return self.wrap_collection(refs, hints=hints)

    @validation.request_body_schema(schema.endpoint_group_create_request_body)
    @validation.response_body_schema(schema.endpoint_group_response_body)
    def post(self):
        """Create new endpoint groups.

        POST /v3/OS-EP-FILTER/endpoint_groups
        """
        ENFORCER.enforce_call(action='identity:create_endpoint_group')
        ep_group = self.request_body_json.get('endpoint_group', {})
        if not ep_group.get('filters'):
            # TODO(morgan): Make this not require substitution. Substitution is
            # done here due to String Freeze in the Rocky release.
            msg = _('%s field is required and cannot be empty') % 'filters'
            raise exception.ValidationError(message=msg)
        _require_valid_filter(ep_group)
        ep_group = self._assign_unique_id(ep_group)
        return (
            self.wrap_member(
                PROVIDERS.catalog_api.create_endpoint_group(
                    ep_group['id'], ep_group
                )
            ),
            http.client.CREATED,
        )


class EndpointGroupResource(ks_flask.ResourceBase):
    collection_key = 'endpoint_groups'
    member_key = 'endpoint_group'
    api_prefix = '/OS-EP-FILTER'

    @validation.request_query_schema(None)
    @validation.response_body_schema(schema.endpoint_group_response_body)
    def get(self, endpoint_group_id=None):
        """Get Endpoint Group

        GET /v3/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}
        """
        ENFORCER.enforce_call(action='identity:get_endpoint_group')
        return self.wrap_member(
            PROVIDERS.catalog_api.get_endpoint_group(endpoint_group_id)
        )

    @validation.request_body_schema(schema.endpoint_group_update_request_body)
    @validation.response_body_schema(schema.endpoint_group_response_body)
    def patch(self, endpoint_group_id):
        """Update existing endpoint groups

        PATCH /v3/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}
        """
        ENFORCER.enforce_call(action='identity:update_endpoint_group')
        ep_group = self.request_body_json.get('endpoint_group', {})
        if 'filters' in ep_group:
            _require_valid_filter(ep_group)
        self._require_matching_id(ep_group)
        return self.wrap_member(
            PROVIDERS.catalog_api.update_endpoint_group(
                endpoint_group_id, ep_group
            )
        )

    def delete(self, endpoint_group_id):
        ENFORCER.enforce_call(action='identity:delete_endpoint_group')
        return (
            PROVIDERS.catalog_api.delete_endpoint_group(endpoint_group_id),
            http.client.NO_CONTENT,
        )


class EPFilterEndpointProjectsResource(flask_restful.Resource):
    def get(self, endpoint_id):
        """Return a list of projects associated with the endpoint."""
        ENFORCER.enforce_call(action='identity:list_projects_for_endpoint')
        PROVIDERS.catalog_api.get_endpoint(endpoint_id)
        refs = PROVIDERS.catalog_api.list_projects_for_endpoint(endpoint_id)
        projects = [
            PROVIDERS.resource_api.get_project(ref['project_id'])
            for ref in refs
        ]
        return ks_flask.ResourceBase.wrap_collection(
            projects, collection_name='projects'
        )


class EPFilterProjectsEndpointsResource(flask_restful.Resource):
    def get(self, project_id, endpoint_id):
        ENFORCER.enforce_call(action='identity:check_endpoint_in_project')
        PROVIDERS.catalog_api.get_endpoint(endpoint_id)
        PROVIDERS.resource_api.get_project(project_id)
        PROVIDERS.catalog_api.check_endpoint_in_project(
            endpoint_id, project_id
        )
        return None, http.client.NO_CONTENT

    def put(self, project_id, endpoint_id):
        ENFORCER.enforce_call(action='identity:add_endpoint_to_project')
        PROVIDERS.catalog_api.get_endpoint(endpoint_id)
        PROVIDERS.resource_api.get_project(project_id)
        PROVIDERS.catalog_api.add_endpoint_to_project(endpoint_id, project_id)
        return None, http.client.NO_CONTENT

    def delete(self, project_id, endpoint_id):
        ENFORCER.enforce_call(action='identity:remove_endpoint_from_project')
        return (
            PROVIDERS.catalog_api.remove_endpoint_from_project(
                endpoint_id, project_id
            ),
            http.client.NO_CONTENT,
        )


class EPFilterProjectEndpointsListResource(flask_restful.Resource):
    def get(self, project_id):
        ENFORCER.enforce_call(action='identity:list_endpoints_for_project')
        PROVIDERS.resource_api.get_project(project_id)
        filtered_endpoints = PROVIDERS.catalog_api.list_endpoints_for_project(
            project_id
        )

        return ks_flask.ResourceBase.wrap_collection(
            [_filter_endpoint(v) for v in filtered_endpoints.values()],
            collection_name='endpoints',
        )


class EndpointFilterProjectEndpointGroupsListResource(flask_restful.Resource):
    def get(self, project_id):
        ENFORCER.enforce_call(
            action='identity:list_endpoint_groups_for_project'
        )
        return EndpointGroupsResource.wrap_collection(
            PROVIDERS.catalog_api.get_endpoint_groups_for_project(project_id)
        )


class EndpointFilterEPGroupsProjects(flask_restful.Resource):
    def get(self, endpoint_group_id):
        ENFORCER.enforce_call(
            action='identity:list_projects_associated_with_endpoint_group'
        )
        endpoint_group_refs = (
            PROVIDERS.catalog_api.list_projects_associated_with_endpoint_group(
                endpoint_group_id
            )
        )
        projects = []
        for endpoint_group_ref in endpoint_group_refs:
            project = PROVIDERS.resource_api.get_project(
                endpoint_group_ref['project_id']
            )
            if project:
                projects.append(project)

        return ks_flask.ResourceBase.wrap_collection(
            projects, collection_name='projects'
        )


class EndpointFilterEPGroupsEndpoints(flask_restful.Resource):
    def get(self, endpoint_group_id):
        ENFORCER.enforce_call(
            action='identity:list_endpoints_associated_with_endpoint_group'
        )
        filtered_endpoints = (
            PROVIDERS.catalog_api.get_endpoints_filtered_by_endpoint_group(
                endpoint_group_id
            )
        )
        return ks_flask.ResourceBase.wrap_collection(
            [_filter_endpoint(e) for e in filtered_endpoints],
            collection_name='endpoints',
        )


class EPFilterGroupsProjectsResource(ks_flask.ResourceBase):
    collection_key = 'project_endpoint_groups'
    member_key = 'project_endpoint_group'

    @classmethod
    def _add_self_referential_link(cls, ref, collection_name=None):
        url = (
            '/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
            '/projects/{project_id}'.format(
                endpoint_group_id=ref['endpoint_group_id'],
                project_id=ref['project_id'],
            )
        )
        ref.setdefault('links', {})
        ref['links']['self'] = url

    def get(self, endpoint_group_id, project_id):
        ENFORCER.enforce_call(action='identity:get_endpoint_group_in_project')
        PROVIDERS.resource_api.get_project(project_id)
        PROVIDERS.catalog_api.get_endpoint_group(endpoint_group_id)
        ref = PROVIDERS.catalog_api.get_endpoint_group_in_project(
            endpoint_group_id, project_id
        )
        return self.wrap_member(ref)

    def put(self, endpoint_group_id, project_id):
        ENFORCER.enforce_call(action='identity:add_endpoint_group_to_project')
        PROVIDERS.resource_api.get_project(project_id)
        PROVIDERS.catalog_api.get_endpoint_group(endpoint_group_id)
        PROVIDERS.catalog_api.add_endpoint_group_to_project(
            endpoint_group_id, project_id
        )
        return None, http.client.NO_CONTENT

    def delete(self, endpoint_group_id, project_id):
        ENFORCER.enforce_call(
            action='identity:remove_endpoint_group_from_project'
        )
        PROVIDERS.resource_api.get_project(project_id)
        PROVIDERS.catalog_api.get_endpoint_group(endpoint_group_id)
        PROVIDERS.catalog_api.remove_endpoint_group_from_project(
            endpoint_group_id, project_id
        )
        return None, http.client.NO_CONTENT


class EPFilterAPI(ks_flask.APIBase):
    _name = 'OS-EP-FILTER'
    _import_name = __name__
    _api_url_prefix = '/OS-EP-FILTER'
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=EndpointGroupsResource,
            url='/endpoint_groups',
            resource_kwargs={},
            rel="endpoint_groups",
            resource_relation_func=_build_resource_relation,
            path_vars=None,
        ),
        ks_flask.construct_resource_map(
            resource=EndpointGroupResource,
            url='/endpoint_groups/<string:endpoint_group_id>',
            resource_kwargs={},
            rel="endpoint_group",
            resource_relation_func=_build_resource_relation,
            path_vars={
                'endpoint_group_id': _ENDPOINT_GROUP_PARAMETER_RELATION
            },
        ),
        ks_flask.construct_resource_map(
            resource=EPFilterEndpointProjectsResource,
            url='/endpoints/<string:endpoint_id>/projects',
            resource_kwargs={},
            rel='endpoint_projects',
            resource_relation_func=_build_resource_relation,
            path_vars={'endpoint_id': json_home.Parameters.ENDPOINT_ID},
        ),
        ks_flask.construct_resource_map(
            resource=EPFilterProjectsEndpointsResource,
            url='/projects/<string:project_id>/endpoints/<string:endpoint_id>',
            resource_kwargs={},
            rel='project_endpoint',
            resource_relation_func=_build_resource_relation,
            path_vars={
                'endpoint_id': json_home.Parameters.ENDPOINT_ID,
                'project_id': json_home.Parameters.PROJECT_ID,
            },
        ),
        ks_flask.construct_resource_map(
            resource=EPFilterProjectEndpointsListResource,
            url='/projects/<string:project_id>/endpoints',
            resource_kwargs={},
            rel='project_endpoints',
            resource_relation_func=_build_resource_relation,
            path_vars={'project_id': json_home.Parameters.PROJECT_ID},
        ),
        ks_flask.construct_resource_map(
            resource=EndpointFilterProjectEndpointGroupsListResource,
            url='/projects/<string:project_id>/endpoint_groups',
            resource_kwargs={},
            rel='project_endpoint_groups',
            resource_relation_func=_build_resource_relation,
            path_vars={'project_id': json_home.Parameters.PROJECT_ID},
        ),
        ks_flask.construct_resource_map(
            resource=EndpointFilterEPGroupsEndpoints,
            url='/endpoint_groups/<string:endpoint_group_id>/endpoints',
            resource_kwargs={},
            rel='endpoints_in_endpoint_group',
            resource_relation_func=_build_resource_relation,
            path_vars={
                'endpoint_group_id': _ENDPOINT_GROUP_PARAMETER_RELATION
            },
        ),
        ks_flask.construct_resource_map(
            resource=EndpointFilterEPGroupsProjects,
            url='/endpoint_groups/<string:endpoint_group_id>/projects',
            resource_kwargs={},
            rel='projects_associated_with_endpoint_group',
            resource_relation_func=_build_resource_relation,
            path_vars={
                'endpoint_group_id': _ENDPOINT_GROUP_PARAMETER_RELATION
            },
        ),
        ks_flask.construct_resource_map(
            resource=EPFilterGroupsProjectsResource,
            url=(
                '/endpoint_groups/<string:endpoint_group_id>/projects/'
                '<string:project_id>'
            ),
            resource_kwargs={},
            rel='endpoint_group_to_project_association',
            resource_relation_func=_build_resource_relation,
            path_vars={
                'project_id': json_home.Parameters.PROJECT_ID,
                'endpoint_group_id': _ENDPOINT_GROUP_PARAMETER_RELATION,
            },
        ),
    ]


APIs = (EPFilterAPI,)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
from urllib import parse
import uuid

import flask
from oslo_log import log
from pycadf import cadftaxonomy as taxonomy

from keystone.auth import plugins as auth_plugins
from keystone.auth.plugins import base
from keystone.common import provider_api
from keystone import exception
from keystone.federation import constants as federation_constants
from keystone.federation import utils
from keystone.i18n import _
from keystone import notifications

LOG = log.getLogger(__name__)

METHOD_NAME = 'mapped'
PROVIDERS = provider_api.ProviderAPIs


class Mapped(base.AuthMethodHandler):
    def _get_token_ref(self, auth_payload):
        token_id = auth_payload['id']
        return PROVIDERS.token_provider_api.validate_token(token_id)

    def authenticate(self, auth_payload):
        """Authenticate mapped user and set an authentication context.

        :param auth_payload: the content of the authentication for a
                             given method

        In addition to ``user_id`` in ``response_data``, this plugin sets
        ``group_ids``, ``OS-FEDERATION:identity_provider`` and
        ``OS-FEDERATION:protocol``

        """
        if 'id' in auth_payload:
            token_ref = self._get_token_ref(auth_payload)
            response_data = handle_scoped_token(
                token_ref, PROVIDERS.federation_api, PROVIDERS.identity_api
            )
        else:
            response_data = handle_unscoped_token(
                auth_payload,
                PROVIDERS.resource_api,
                PROVIDERS.federation_api,
                PROVIDERS.identity_api,
                PROVIDERS.assignment_api,
                PROVIDERS.role_api,
            )

        return base.AuthHandlerResponse(
            status=True, response_body=None, response_data=response_data
        )


def handle_scoped_token(token, federation_api, identity_api):
    response_data = {}
    utils.validate_expiration(token)
    token_audit_id = token.audit_id
    identity_provider = token.identity_provider_id
    protocol = token.protocol_id
    user_id = token.user_id
    group_ids = []
    for group_dict in token.federated_groups:
        group_ids.append(group_dict['id'])
    send_notification = functools.partial(
        notifications.send_saml_audit_notification,
        'authenticate',
        user_id,
        group_ids,
        identity_provider,
        protocol,
        token_audit_id,
    )

    utils.assert_enabled_identity_provider(federation_api, identity_provider)

    try:
        mapping = federation_api.get_mapping_from_idp_and_protocol(
            identity_provider, protocol
        )
        utils.validate_mapped_group_ids(group_ids, mapping['id'], identity_api)

    except Exception:
        # NOTE(topol): Diaper defense to catch any exception, so we can
        # send off failed authentication notification, raise the exception
        # after sending the notification
        send_notification(taxonomy.OUTCOME_FAILURE)
        raise
    else:
        send_notification(taxonomy.OUTCOME_SUCCESS)

    response_data['user_id'] = user_id
    response_data['group_ids'] = group_ids
    response_data[federation_constants.IDENTITY_PROVIDER] = identity_provider
    response_data[federation_constants.PROTOCOL] = protocol

    return response_data


def configure_project_domain(shadow_project, idp_domain_id, resource_api):
    """Configure federated projects domain.

    We set the domain to be the default (idp_domain_id) if the project
    from the attribute mapping comes without a domain.
    """
    LOG.debug('Processing domain for project: %s', shadow_project)
    domain = shadow_project.get('domain', {"id": idp_domain_id})
    if 'id' not in domain:
        db_domain = resource_api.get_domain_by_name(domain['name'])
        domain = {"id": db_domain.get('id')}
    shadow_project['domain'] = domain
    LOG.debug(
        'Project [%s] domain ID was resolved to [%s]',
        shadow_project['name'],
        shadow_project['domain']['id'],
    )


def handle_projects_from_mapping(
    shadow_projects,
    idp_domain_id,
    existing_roles,
    user,
    assignment_api,
    resource_api,
):
    for shadow_project in shadow_projects:
        configure_project_domain(shadow_project, idp_domain_id, resource_api)
        try:
            # Check and see if the project already exists and if it
            # does not, try to create it.
            project = resource_api.get_project_by_name(
                shadow_project['name'], shadow_project['domain']['id']
            )
        except exception.ProjectNotFound:
            LOG.info(
                'Project %(project_name)s does not exist. It will be '
                'automatically provisioning for user %(user_id)s.',
                {
                    'project_name': shadow_project['name'],
                    'user_id': user['id'],
                },
            )
            project_ref = {
                'id': uuid.uuid4().hex,
                'name': shadow_project['name'],
                'domain_id': shadow_project['domain']['id'],
            }
            project = resource_api.create_project(
                project_ref['id'], project_ref
            )
        shadow_roles = shadow_project['roles']
        for shadow_role in shadow_roles:
            assignment_api.create_grant(
                existing_roles[shadow_role['name']]['id'],
                user_id=user['id'],
                project_id=project['id'],
            )


def handle_unscoped_token(
    auth_payload,
    resource_api,
    federation_api,
    identity_api,
    assignment_api,
    role_api,
):
    def validate_shadow_mapping(
        shadow_projects, existing_roles, user_domain_id, idp_id
    ):
        # Validate that the roles in the shadow mapping actually exist. If
        # they don't we should bail early before creating anything.
        for shadow_project in shadow_projects:
            for shadow_role in shadow_project['roles']:
                # The role in the project mapping must exist in order for it to
                # be useful.
                if shadow_role['name'] not in existing_roles:
                    LOG.error(
                        'Role %s was specified in the mapping but does '
                        'not exist. All roles specified in a mapping must '
                        'exist before assignment.',
                        shadow_role['name'],
                    )
                    # NOTE(lbragstad): The RoleNotFound exception usually
                    # expects a role_id as the parameter, but in this case we
                    # only have a name so we'll pass that instead.
                    raise exception.RoleNotFound(shadow_role['name'])
                role = existing_roles[shadow_role['name']]
                if (
                    role['domain_id'] is not None
                    and role['domain_id'] != user_domain_id
                ):
                    LOG.error(
                        'Role %(role)s is a domain-specific role and '
                        'cannot be assigned within %(domain)s.',
                        {
                            'role': shadow_role['name'],
                            'domain': user_domain_id,
                        },
                    )
                    raise exception.DomainSpecificRoleNotWithinIdPDomain(
                        role_name=shadow_role['name'], identity_provider=idp_id
                    )

    def is_ephemeral_user(mapped_properties):
        return mapped_properties['user']['type'] == utils.UserType.EPHEMERAL

    def build_ephemeral_user_context(
        user, mapped_properties, identity_provider, protocol
    ):
        resp = {}
        resp['user_id'] = user['id']
        resp['group_ids'] = mapped_properties['group_ids']
        resp[federation_constants.IDENTITY_PROVIDER] = identity_provider
        resp[federation_constants.PROTOCOL] = protocol

        return resp

    def build_local_user_context(mapped_properties):
        resp = {}
        user_info = auth_plugins.UserAuthInfo.create(
            mapped_properties, METHOD_NAME
        )
        resp['user_id'] = user_info.user_id

        return resp

    assertion = extract_assertion_data()
    try:
        identity_provider = auth_payload['identity_provider']
    except KeyError:
        raise exception.ValidationError(
            attribute='identity_provider', target='mapped'
        )
    try:
        protocol = auth_payload['protocol']
    except KeyError:
        raise exception.ValidationError(attribute='protocol', target='mapped')

    utils.assert_enabled_identity_provider(federation_api, identity_provider)

    group_ids = None
    # NOTE(topol): The user is coming in from an IdP with a SAML assertion
    # instead of from a token, so we set token_id to None
    token_id = None
    # NOTE(marek-denis): This variable is set to None and there is a
    # possibility that it will be used in the CADF notification. This means
    # operation will not be mapped to any user (even ephemeral).
    user_id = None

    try:
        try:
            mapped_properties, mapping_id = apply_mapping_filter(
                identity_provider,
                protocol,
                assertion,
                resource_api,
                federation_api,
                identity_api,
            )
        except exception.ValidationError as e:
            # if mapping is either invalid or yield no valid identity,
            # it is considered a failed authentication
            raise exception.Unauthorized(e)

        if is_ephemeral_user(mapped_properties):
            idp_domain_id = federation_api.get_idp(identity_provider)[
                'domain_id'
            ]

            validate_and_prepare_federated_user(
                mapped_properties, idp_domain_id, resource_api
            )

            user = identity_api.shadow_federated_user(
                identity_provider,
                protocol,
                mapped_properties['user'],
                group_ids=mapped_properties['group_ids'],
            )

            if 'projects' in mapped_properties:
                existing_roles = {
                    role['name']: role for role in role_api.list_roles()
                }
                # NOTE(lbragstad): If we are dealing with a shadow mapping,
                # then we need to make sure we validate all pieces of the
                # mapping and what it's saying to create. If there is something
                # wrong with how the mapping is, we should bail early before we
                # create anything.

                validate_shadow_mapping(
                    mapped_properties['projects'],
                    existing_roles,
                    mapped_properties['user']['domain']['id'],
                    identity_provider,
                )
                handle_projects_from_mapping(
                    mapped_properties['projects'],
                    idp_domain_id,
                    existing_roles,
                    user,
                    assignment_api,
                    resource_api,
                )

            user_id = user['id']
            group_ids = mapped_properties['group_ids']
            response_data = build_ephemeral_user_context(
                user, mapped_properties, identity_provider, protocol
            )
        else:
            response_data = build_local_user_context(mapped_properties)

    except Exception:
        # NOTE(topol): Diaper defense to catch any exception, so we can
        # send off failed authentication notification, raise the exception
        # after sending the notification
        outcome = taxonomy.OUTCOME_FAILURE
        notifications.send_saml_audit_notification(
            'authenticate',
            user_id,
            group_ids,
            identity_provider,
            protocol,
            token_id,
            outcome,
        )
        raise
    else:
        outcome = taxonomy.OUTCOME_SUCCESS
        notifications.send_saml_audit_notification(
            'authenticate',
            user_id,
            group_ids,
            identity_provider,
            protocol,
            token_id,
            outcome,
        )

    return response_data


def extract_assertion_data():
    assertion = dict(utils.get_assertion_params_from_env())
    return assertion


def apply_mapping_filter(
    identity_provider,
    protocol,
    assertion,
    resource_api,
    federation_api,
    identity_api,
):
    idp = federation_api.get_idp(identity_provider)
    utils.validate_idp(idp, protocol, assertion)

    mapped_properties, mapping_id = federation_api.evaluate(
        identity_provider, protocol, assertion
    )

    # NOTE(marek-denis): We update group_ids only here to avoid fetching
    # groups identified by name/domain twice.
    # NOTE(marek-denis): Groups are translated from name/domain to their
    # corresponding ids in the auth plugin, as we need information what
    # ``mapping_id`` was used as well as idenity_api and resource_api
    # objects.
    group_ids = mapped_properties['group_ids']
    utils.validate_mapped_group_ids(group_ids, mapping_id, identity_api)
    group_ids.extend(
        utils.transform_to_group_ids(
            mapped_properties['group_names'],
            mapping_id,
            identity_api,
            resource_api,
        )
    )
    mapped_properties['group_ids'] = list(set(group_ids))
    return mapped_properties, mapping_id


def validate_and_prepare_federated_user(
    mapped_properties, idp_domain_id, resource_api
):
    """Setup federated username.

    Function covers all the cases for properly setting user id, a primary
    identifier for identity objects. Initial version of the mapping engine
    assumed user is identified by ``name`` and his ``id`` is built from the
    name. We, however need to be able to accept local rules that identify user
    by either id or name/domain.

    The following use-cases are covered:

    1) If neither user_name nor user_id is set raise exception.Unauthorized
    2) If user_id is set and user_name not, set user_name equal to user_id
    3) If user_id is not set and user_name is, set user_id as url safe version
       of user_name.

    Furthermore, we set the IdP as the user domain, if the user definition
    does not come with a domain definition.

    :param mapped_properties: Properties issued by a RuleProcessor.
    :type: dictionary

    :param idp_domain_id: The domain ID of the IdP registered in OpenStack.
    :type: string

    :param resource_api: The resource API used to access the database layer.
    :type: object

    :raises keystone.exception.Unauthorized: If neither `user_name` nor
        `user_id` is set.
    :returns: tuple with user identification
    :rtype: tuple

    """
    user = mapped_properties['user']

    user_id = user.get('id')
    user_name = user.get('name') or flask.request.remote_user

    if not any([user_id, user_name]):
        msg = _(
            "Could not map user while setting ephemeral user identity. "
            "Either mapping rules must specify user id/name or "
            "REMOTE_USER environment variable must be set."
        )
        raise exception.Unauthorized(msg)

    elif not user_name:
        user['name'] = user_id

    elif not user_id:
        user_id = user_name

    if user_name:
        user['name'] = user_name
    user['id'] = parse.quote(user_id)

    LOG.debug('Processing domain for federated user: %s', user)
    domain = user.get('domain', {"id": idp_domain_id})
    if 'id' not in domain:
        db_domain = resource_api.get_domain_by_name(domain['name'])
        domain = {"id": db_domain.get('id')}

    user['domain'] = domain
    LOG.debug(
        'User [%s] domain ID was resolved to [%s]',
        user['name'],
        user['domain']['id'],
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

allow_redelegation = cfg.BoolOpt(
    'allow_redelegation',
    default=False,
    help=utils.fmt(
        """
Allows authorization to be redelegated from one user to another, effectively
chaining trusts together. When disabled, the `remaining_uses` attribute of a
trust is constrained to be zero.
"""
    ),
)

max_redelegation_count = cfg.IntOpt(
    'max_redelegation_count',
    default=3,
    help=utils.fmt(
        """
Maximum number of times that authorization can be redelegated from one user to
another in a chain of trusts. This number may be reduced further for a specific
trust.
"""
    ),
)

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt(
        """
Entry point for the trust backend driver in the `keystone.trust` namespace.
Keystone only provides a `sql` driver, so there is no reason to change this
unless you are providing a custom entry point.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [allow_redelegation, max_redelegation_count, driver]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import requests
import testtools

from keystone.tests.common import auth as common_auth


class BaseTestCase(testtools.TestCase, common_auth.AuthTestMixin):
    request_headers = {'content-type': 'application/json'}

    def setUp(self):
        self.ADMIN_URL = os.environ.get(
            'KSTEST_ADMIN_URL', 'http://localhost:5000'
        )
        self.PUBLIC_URL = os.environ.get(
            'KSTEST_PUBLIC_URL', 'http://localhost:5000'
        )
        self.admin = {
            'name': os.environ.get('KSTEST_ADMIN_USERNAME', 'admin'),
            'password': os.environ.get('KSTEST_ADMIN_PASSWORD', ''),
            'domain_id': os.environ.get('KSTEST_ADMIN_DOMAIN_ID', 'default'),
        }

        self.user = {
            'name': os.environ.get('KSTEST_USER_USERNAME', 'demo'),
            'password': os.environ.get('KSTEST_USER_PASSWORD', ''),
            'domain_id': os.environ.get('KSTEST_USER_DOMAIN_ID', 'default'),
        }

        self.project_id = os.environ.get('KSTEST_PROJECT_ID')
        self.project_name = os.environ.get('KSTEST_PROJECT_NAME')
        self.project_domain_id = os.environ.get('KSTEST_PROJECT_DOMAIN_ID')

        super().setUp()

    def _http_headers(self, token=None):
        headers = {'content-type': 'application/json'}
        if token:
            headers['X-Auth-Token'] = token
        return headers

    def get_scoped_token_response(self, user):
        """Convenience method so that we can test authenticated requests.

        :param user: A dictionary with user information like 'username',
                     'password', 'domain_id'
        :returns: urllib3.Response object

        """
        body = self.build_authentication_request(
            username=user['name'],
            user_domain_name=user['domain_id'],
            password=user['password'],
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
        )
        return requests.post(
            self.PUBLIC_URL + '/v3/auth/tokens',
            headers=self.request_headers,
            json=body,
        )

    def get_scoped_token(self, user):
        """Convenience method for getting scoped token.

        This method doesn't do any token validation.

        :param user: A dictionary with user information like 'username',
                     'password', 'domain_id'
        :returns: An OpenStack token for further use
        :rtype: str

        """
        r = self.get_scoped_token_response(user)
        return r.headers.get('X-Subject-Token')

    def get_scoped_admin_token(self):
        return self.get_scoped_token(self.admin)

    def get_scoped_user_token(self):
        return self.get_scoped_token(self.user)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import typing as ty
import uuid

import fixtures
import flask
import flask_restful
from oslo_policy import policy
from oslo_serialization import jsonutils
from testtools import matchers

from keystone.common import context
from keystone.common import json_home
from keystone.common import rbac_enforcer
import keystone.conf
from keystone import exception
from keystone.server.flask import common as flask_common
from keystone.server.flask.request_processing import json_body
from keystone.tests.unit import rest

CONF = keystone.conf.CONF


class _TestResourceWithCollectionInfo(flask_common.ResourceBase):
    collection_key = 'arguments'
    member_key = 'argument'
    __shared_state__: dict[str, ty.Any] = {}
    _storage_dict: dict[str, ty.Any] = {}

    def __init__(self):
        super().__init__()
        # Share State, this is for "dummy" backend storage.
        self.__dict__ = self.__shared_state__

    @classmethod
    def _reset(cls):
        # Used after a test to ensure clean-state
        cls._storage_dict.clear()
        cls.__shared_state__.clear()

    def _list_arguments(self):
        return self.wrap_collection(list(self._storage_dict.values()))

    def get(self, argument_id=None):
        # List with no argument, get resource with id, used for HEAD as well.
        rbac_enforcer.enforcer.RBACEnforcer.enforce_call(
            action='example:allowed'
        )
        if argument_id is None:
            # List
            return self._list_arguments()
        else:
            # get resource with id
            try:
                return self.wrap_member(self._storage_dict[argument_id])
            except KeyError:
                raise exception.NotFound(target=argument_id)

    def post(self):
        rbac_enforcer.enforcer.RBACEnforcer.enforce_call(
            action='example:allowed'
        )
        ref = flask.request.get_json(force=True)
        ref = self._assign_unique_id(ref)
        self._storage_dict[ref['id']] = ref
        return self.wrap_member(self._storage_dict[ref['id']]), 201

    def put(self, argument_id):
        rbac_enforcer.enforcer.RBACEnforcer.enforce_call(
            action='example:allowed'
        )
        try:
            self._storage_dict[argument_id]
        except KeyError:
            raise exception.NotFound(target=argument_id)
        ref = flask.request.get_json(force=True)
        self._require_matching_id(ref)
        # Maintain the ref id
        ref['id'] = argument_id
        self._storage_dict[argument_id] = ref
        return '', 204

    def patch(self, argument_id):
        rbac_enforcer.enforcer.RBACEnforcer.enforce_call(
            action='example:allowed'
        )
        try:
            self._storage_dict[argument_id]
        except KeyError:
            raise exception.NotFound(target=argument_id)
        ref = flask.request.get_json(force=True)
        self._require_matching_id(ref)
        self._storage_dict[argument_id].update(ref)
        return self.wrap_member(self._storage_dict[argument_id])

    def delete(self, argument_id):
        rbac_enforcer.enforcer.RBACEnforcer.enforce_call(
            action='example:allowed'
        )
        try:
            del self._storage_dict[argument_id]
        except KeyError:
            raise exception.NotFound(target=argument_id)
        return '', 204


class _TestRestfulAPI(flask_common.APIBase):
    _name = 'test_api_base'
    _import_name = __name__
    resources = []
    resource_mapping = []

    def __init__(self, *args, **kwargs):
        self.resource_mapping = kwargs.pop('resource_mapping', [])
        self.resources = kwargs.pop(
            'resources', [_TestResourceWithCollectionInfo]
        )
        super().__init__(*args, **kwargs)


class TestKeystoneFlaskCommon(rest.RestfulTestCase):
    _policy_rules = [
        policy.RuleDefault(name='example:allowed', check_str=''),
        policy.RuleDefault(name='example:deny', check_str='false:false'),
    ]

    def setUp(self):
        super().setUp()
        enf = rbac_enforcer.enforcer.RBACEnforcer()

        def register_rules(enf_obj):
            enf_obj.register_defaults(self._policy_rules)

        self.useFixture(
            fixtures.MockPatchObject(enf, 'register_rules', register_rules)
        )
        self.useFixture(
            fixtures.MockPatchObject(
                rbac_enforcer.enforcer,
                '_POSSIBLE_TARGET_ACTIONS',
                {r.name for r in self._policy_rules},
            )
        )

        enf._reset()
        self.addCleanup(enf._reset)
        self.addCleanup(_TestResourceWithCollectionInfo._reset)

    def _get_token(self):
        auth_json = {
            'auth': {
                'identity': {
                    'methods': ['password'],
                    'password': {
                        'user': {
                            'name': self.user_req_admin['name'],
                            'password': self.user_req_admin['password'],
                            'domain': {'id': self.user_req_admin['domain_id']},
                        }
                    },
                },
                'scope': {'project': {'id': self.project_service['id']}},
            }
        }
        return (
            self.test_client()
            .post('/v3/auth/tokens', json=auth_json, expected_status_code=201)
            .headers['X-Subject-Token']
        )

    def _setup_flask_restful_api(self, **options):
        self.restful_api_opts = options.copy()
        orig_value = _TestResourceWithCollectionInfo.api_prefix
        setattr(
            _TestResourceWithCollectionInfo,
            'api_prefix',
            options.get('api_url_prefix', ''),
        )
        self.addCleanup(
            setattr, _TestResourceWithCollectionInfo, 'api_prefix', orig_value
        )
        self.restful_api = _TestRestfulAPI(**options)
        self.public_app.app.register_blueprint(self.restful_api.blueprint)
        self.cleanup_instance('restful_api')
        self.cleanup_instance('restful_api_opts')

    def _make_requests(self):
        path_base = '/arguments'
        api_prefix = self.restful_api_opts.get('api_url_prefix', '')
        blueprint_prefix = self.restful_api._blueprint_url_prefix.rstrip('/')
        url = ''.join(
            [x for x in [blueprint_prefix, api_prefix, path_base] if x]
        )
        headers = {'X-Auth-Token': self._get_token()}
        with self.test_client() as c:
            # GET LIST
            resp = c.get(url, headers=headers)
            self.assertEqual(
                _TestResourceWithCollectionInfo.wrap_collection([]), resp.json
            )
            unknown_id = uuid.uuid4().hex

            # GET non-existent ref
            c.get(
                f'{url}/{unknown_id}',
                headers=headers,
                expected_status_code=404,
            )

            # HEAD non-existent ref
            c.head(
                f'{url}/{unknown_id}',
                headers=headers,
                expected_status_code=404,
            )

            # PUT non-existent ref
            c.put(
                f'{url}/{unknown_id}',
                json={},
                headers=headers,
                expected_status_code=404,
            )

            # PATCH non-existent ref
            c.patch(
                f'{url}/{unknown_id}',
                json={},
                headers=headers,
                expected_status_code=404,
            )

            # DELETE non-existent ref
            c.delete(
                f'{url}/{unknown_id}',
                headers=headers,
                expected_status_code=404,
            )

            # POST new ref
            new_argument_resource = {'testing': uuid.uuid4().hex}
            new_argument_resp = c.post(
                url, json=new_argument_resource, headers=headers
            ).json['argument']

            # POST second new ref
            new_argument2_resource = {'testing': uuid.uuid4().hex}
            new_argument2_resp = c.post(
                url, json=new_argument2_resource, headers=headers
            ).json['argument']

            # GET list
            get_list_resp = c.get(url, headers=headers).json
            self.assertIn(new_argument_resp, get_list_resp['arguments'])
            self.assertIn(new_argument2_resp, get_list_resp['arguments'])

            # GET first ref
            get_resp = c.get(
                '{}/{}'.format(url, new_argument_resp['id']), headers=headers
            ).json['argument']
            self.assertEqual(new_argument_resp, get_resp)

            # HEAD first ref
            head_resp = c.head(
                '{}/{}'.format(url, new_argument_resp['id']), headers=headers
            ).data
            # NOTE(morgan): For python3 compat, explicitly binary type
            self.assertEqual(head_resp, b'')

            # PUT update first ref
            replacement_argument = {'new_arg': True, 'id': uuid.uuid4().hex}
            c.put(
                '{}/{}'.format(url, new_argument_resp['id']),
                headers=headers,
                json=replacement_argument,
                expected_status_code=400,
            )
            replacement_argument.pop('id')
            c.put(
                '{}/{}'.format(url, new_argument_resp['id']),
                headers=headers,
                json=replacement_argument,
            )
            put_resp = c.get(
                '{}/{}'.format(url, new_argument_resp['id']), headers=headers
            ).json['argument']
            self.assertNotIn(new_argument_resp['testing'], put_resp)
            self.assertTrue(put_resp['new_arg'])

            # GET first ref (check for replacement)
            get_replacement_resp = c.get(
                '{}/{}'.format(url, new_argument_resp['id']), headers=headers
            ).json['argument']
            self.assertEqual(put_resp, get_replacement_resp)

            # PATCH update first ref
            patch_ref = {'uuid': uuid.uuid4().hex}
            patch_resp = c.patch(
                '{}/{}'.format(url, new_argument_resp['id']),
                headers=headers,
                json=patch_ref,
            ).json['argument']
            self.assertTrue(patch_resp['new_arg'])
            self.assertEqual(patch_ref['uuid'], patch_resp['uuid'])

            # GET first ref (check for update)
            get_patched_ref_resp = c.get(
                '{}/{}'.format(url, new_argument_resp['id']), headers=headers
            ).json['argument']
            self.assertEqual(patch_resp, get_patched_ref_resp)

            # DELETE first ref
            c.delete(
                '{}/{}'.format(url, new_argument_resp['id']), headers=headers
            )
            # Check that it was in-fact deleted
            c.get(
                '{}/{}'.format(url, new_argument_resp['id']),
                headers=headers,
                expected_status_code=404,
            )

    def test_api_url_prefix(self):
        url_prefix = f'/{uuid.uuid4().hex}'
        self._setup_flask_restful_api(api_url_prefix=url_prefix)
        self._make_requests()

    def test_blueprint_url_prefix(self):
        url_prefix = f'/{uuid.uuid4().hex}'
        self._setup_flask_restful_api(blueprint_url_prefix=url_prefix)
        self._make_requests()

    def test_build_restful_api_no_prefix(self):
        self._setup_flask_restful_api()
        self._make_requests()

    def test_cannot_add_before_request_functions_twice(self):
        class TestAPIDuplicateBefore(_TestRestfulAPI):
            def __init__(self):
                super().__init__()
                self._register_before_request_functions()

        self.assertRaises(AssertionError, TestAPIDuplicateBefore)

    def test_cannot_add_after_request_functions_twice(self):
        class TestAPIDuplicateAfter(_TestRestfulAPI):
            def __init__(self):
                super().__init__()
                self._register_after_request_functions()

        self.assertRaises(AssertionError, TestAPIDuplicateAfter)

    def test_after_request_functions_must_be_added(self):
        class TestAPINoAfter(_TestRestfulAPI):
            def _register_after_request_functions(self, functions=None):
                pass

        self.assertRaises(AssertionError, TestAPINoAfter)

    def test_before_request_functions_must_be_added(self):
        class TestAPINoBefore(_TestRestfulAPI):
            def _register_before_request_functions(self, functions=None):
                pass

        self.assertRaises(AssertionError, TestAPINoBefore)

    def test_before_request_functions(self):
        # Test additional "before" request functions fire.
        attr = uuid.uuid4().hex

        def do_something():
            setattr(flask.g, attr, True)

        class TestAPI(_TestRestfulAPI):
            def _register_before_request_functions(self, functions=None):
                functions = functions or []
                functions.append(do_something)
                super()._register_before_request_functions(functions)

        api = TestAPI(resources=[_TestResourceWithCollectionInfo])
        self.public_app.app.register_blueprint(api.blueprint)
        token = self._get_token()
        with self.test_client() as c:
            c.get('/v3/arguments', headers={'X-Auth-Token': token})
            self.assertTrue(getattr(flask.g, attr, False))

    def test_after_request_functions(self):
        # Test additional "after" request functions fire. In this case, we
        # alter the response code to 420
        attr = uuid.uuid4().hex

        def do_something(resp):
            setattr(flask.g, attr, True)
            resp.status_code = 420
            return resp

        class TestAPI(_TestRestfulAPI):
            def _register_after_request_functions(self, functions=None):
                functions = functions or []
                functions.append(do_something)
                super()._register_after_request_functions(functions)

        api = TestAPI(resources=[_TestResourceWithCollectionInfo])
        self.public_app.app.register_blueprint(api.blueprint)
        token = self._get_token()
        with self.test_client() as c:
            c.get(
                '/v3/arguments',
                headers={'X-Auth-Token': token},
                expected_status_code=420,
            )

    def test_construct_resource_map(self):
        resource_name = 'arguments'
        param_relation = json_home.build_v3_parameter_relation('argument_id')
        alt_rel_func = functools.partial(
            json_home.build_v3_extension_resource_relation,
            extension_name='extension',
            extension_version='1.0',
        )
        url = '/v3/arguments/<string:argument_id>'
        old_url = [
            {
                'url': '/v3/old_arguments/<string:argument_id>',
                'json_home': flask_common.construct_json_home_data(
                    rel='arguments', resource_relation_func=alt_rel_func
                ),
            }
        ]

        mapping = flask_common.construct_resource_map(
            resource=_TestResourceWithCollectionInfo,
            url=url,
            resource_kwargs={},
            alternate_urls=old_url,
            rel=resource_name,
            status=json_home.Status.EXPERIMENTAL,
            path_vars={'argument_id': param_relation},
            resource_relation_func=json_home.build_v3_resource_relation,
        )
        self.assertEqual(_TestResourceWithCollectionInfo, mapping.resource)
        self.assertEqual(url, mapping.url)
        self.assertEqual(
            json_home.build_v3_resource_relation(resource_name),
            mapping.json_home_data.rel,
        )
        self.assertEqual(
            json_home.Status.EXPERIMENTAL, mapping.json_home_data.status
        )
        self.assertEqual(
            {'argument_id': param_relation}, mapping.json_home_data.path_vars
        )
        # Check the alternate URL data is populated sanely
        self.assertEqual(1, len(mapping.alternate_urls))
        alt_url_data = mapping.alternate_urls[0]
        self.assertEqual(old_url[0]['url'], alt_url_data['url'])
        self.assertEqual(old_url[0]['json_home'], alt_url_data['json_home'])

    def test_instantiate_and_register_to_app(self):
        # Test that automatic instantiation and registration to app works.
        self.restful_api_opts = {}
        self.restful_api = _TestRestfulAPI.instantiate_and_register_to_app(
            self.public_app.app
        )
        self.cleanup_instance('restful_api_opts')
        self.cleanup_instance('restful_api')
        self._make_requests()

    def test_unenforced_api_decorator(self):
        # Test unenforced decorator works as expected

        class MappedResource(flask_restful.Resource):
            @flask_common.unenforced_api
            def post(self):
                post_body = flask.request.get_json()
                return {'post_body': post_body}, 201

        resource_map = flask_common.construct_resource_map(
            resource=MappedResource,
            url='test_api',
            alternate_urls=[],
            resource_kwargs={},
            rel='test',
            status=json_home.Status.STABLE,
            path_vars=None,
            resource_relation_func=json_home.build_v3_resource_relation,
        )

        restful_api = _TestRestfulAPI(
            resource_mapping=[resource_map], resources=[]
        )
        self.public_app.app.register_blueprint(restful_api.blueprint)
        token = self._get_token()
        with self.test_client() as c:
            body = {'test_value': uuid.uuid4().hex}
            # Works with token
            resp = c.post(
                '/v3/test_api', json=body, headers={'X-Auth-Token': token}
            )
            self.assertEqual(body, resp.json['post_body'])
            # Works without token
            resp = c.post('/v3/test_api', json=body)
            self.assertEqual(body, resp.json['post_body'])

    def test_HTTP_OPTIONS_is_unenforced(self):
        # Standup a test mapped resource and call OPTIONS on it. This will
        # return a header "Allow" with the valid methods, in this case
        # OPTIONS and POST. Ensure that the response otherwise conforms
        # as expected.

        class MappedResource(flask_restful.Resource):
            def post(self):
                # we don't actually use this or call it.
                pass

        resource_map = flask_common.construct_resource_map(
            resource=MappedResource,
            url='test_api',
            alternate_urls=[],
            resource_kwargs={},
            rel='test',
            status=json_home.Status.STABLE,
            path_vars=None,
            resource_relation_func=json_home.build_v3_resource_relation,
        )

        restful_api = _TestRestfulAPI(
            resource_mapping=[resource_map], resources=[]
        )
        self.public_app.app.register_blueprint(restful_api.blueprint)
        with self.test_client() as c:
            r = c.options('/v3/test_api')
            # make sure we split the data and left/right strip off whitespace
            # The use of a SET here is to ensure the exact values are in place
            # even if hash-seeds change order. `r.data` will be an empty
            # byte-string. `Content-Length` will be 0.
            self.assertEqual(
                {'OPTIONS', 'POST'},
                {v.lstrip().rstrip() for v in r.headers['Allow'].split(',')},
            )
            self.assertEqual(r.headers['Content-Length'], '0')
            self.assertEqual(r.data, b'')

    def test_mapped_resource_routes(self):
        # Test non-standard URL routes ("mapped") function as expected

        class MappedResource(flask_restful.Resource):
            def post(self):
                rbac_enforcer.enforcer.RBACEnforcer().enforce_call(
                    action='example:allowed'
                )
                post_body = flask.request.get_json()
                return {'post_body': post_body}, 201

        resource_map = flask_common.construct_resource_map(
            resource=MappedResource,
            url='test_api',
            alternate_urls=[],
            resource_kwargs={},
            rel='test',
            status=json_home.Status.STABLE,
            path_vars=None,
            resource_relation_func=json_home.build_v3_resource_relation,
        )

        restful_api = _TestRestfulAPI(
            resource_mapping=[resource_map], resources=[]
        )
        self.public_app.app.register_blueprint(restful_api.blueprint)
        token = self._get_token()
        with self.test_client() as c:
            body = {'test_value': uuid.uuid4().hex}
            resp = c.post(
                '/v3/test_api', json=body, headers={'X-Auth-Token': token}
            )
            self.assertEqual(body, resp.json['post_body'])

    def test_correct_json_home_document(self):
        class MappedResource(flask_restful.Resource):
            def post(self):
                rbac_enforcer.enforcer.RBACEnforcer().enforce_call(
                    action='example:allowed'
                )
                post_body = flask.request.get_json()
                return {'post_body': post_body}

        # NOTE(morgan): totally fabricated json_home data based upon our TEST
        # restful_apis.
        json_home_data = {
            'https://docs.openstack.org/api/openstack-identity/3/'
            'rel/argument': {
                'href-template': '/v3/arguments/{argument_id}',
                'href-vars': {
                    'argument_id': 'https://docs.openstack.org/api/'
                    'openstack-identity/3/param/argument_id'
                },
            },
            'https://docs.openstack.org/api/openstack-identity/3/'
            'rel/arguments': {'href': '/v3/arguments'},
            'https://docs.openstack.org/api/openstack-identity/3/'
            'rel/test': {'href': '/v3/test_api'},
        }

        resource_map = flask_common.construct_resource_map(
            resource=MappedResource,
            url='test_api',
            alternate_urls=[],
            resource_kwargs={},
            rel='test',
            status=json_home.Status.STABLE,
            path_vars=None,
            resource_relation_func=json_home.build_v3_resource_relation,
        )

        restful_api = _TestRestfulAPI(resource_mapping=[resource_map])
        self.public_app.app.register_blueprint(restful_api.blueprint)

        with self.test_client() as c:
            headers = {'Accept': 'application/json-home'}
            resp = c.get('/', headers=headers)
            resp_data = jsonutils.loads(resp.data)
            for rel in json_home_data:
                self.assertThat(
                    resp_data['resources'][rel],
                    matchers.Equals(json_home_data[rel]),
                )

    def test_normalize_domain_id_extracts_domain_id_if_needed(self):
        self._setup_flask_restful_api()
        blueprint_prefix = self.restful_api._blueprint_url_prefix.rstrip('/')
        url = ''.join([blueprint_prefix, '/arguments'])
        headers = {'X-Auth-Token': self._get_token()}
        ref_with_domain_id = {'domain_id': uuid.uuid4().hex}
        ref_without_domain_id = {}
        with self.test_client() as c:
            # Make a dummy request.. ANY request is fine to push the whole
            # context stack.
            c.get(
                f'{url}/{uuid.uuid4().hex}',
                headers=headers,
                expected_status_code=404,
            )

            oslo_context = flask.request.environ[context.REQUEST_CONTEXT_ENV]

            # Normal Project Scope Form
            # ---------------------------
            # test that _normalize_domain_id does something sane
            domain_id = ref_with_domain_id['domain_id']
            # Ensure we don't change the domain if it is specified
            flask_common.ResourceBase._normalize_domain_id(ref_with_domain_id)
            self.assertEqual(domain_id, ref_with_domain_id['domain_id'])
            # Ensure (deprecated) we add default domain if needed
            flask_common.ResourceBase._normalize_domain_id(
                ref_without_domain_id
            )
            self.assertEqual(
                CONF.identity.default_domain_id,
                ref_without_domain_id['domain_id'],
            )
            ref_without_domain_id.clear()

            # Domain Scoped Form
            # --------------------
            # Just set oslo_context domain_id to a value. This is how we
            # communicate domain scope. No need to explicitly
            # do a domain-scoped request, this is a synthetic text anyway
            oslo_context.domain_id = uuid.uuid4().hex
            # Ensure we don't change the domain if it is specified
            flask_common.ResourceBase._normalize_domain_id(ref_with_domain_id)
            self.assertEqual(domain_id, ref_with_domain_id['domain_id'])
            flask_common.ResourceBase._normalize_domain_id(
                ref_without_domain_id
            )
            self.assertEqual(
                oslo_context.domain_id, ref_without_domain_id['domain_id']
            )
            ref_without_domain_id.clear()

            # "Admin" Token form
            # -------------------
            # Explicitly set "is_admin" to true, no new request is needed
            # as we simply check "is_admin" value everywhere
            oslo_context.is_admin = True
            oslo_context.domain_id = None
            # Ensure we don't change the domain if it is specified
            flask_common.ResourceBase._normalize_domain_id(ref_with_domain_id)
            self.assertEqual(domain_id, ref_with_domain_id['domain_id'])
            # Ensure we raise an appropriate exception with the inferred
            # domain_id
            self.assertRaises(
                exception.ValidationError,
                flask_common.ResourceBase._normalize_domain_id,
                ref=ref_without_domain_id,
            )

    def test_api_prefix_self_referential_link_substitution(self):
        view_arg = uuid.uuid4().hex

        class TestResource(flask_common.ResourceBase):
            api_prefix = '/<string:test_value>/nothing'

        # use a dummy request context, no enforcement is happening
        # therefore we don't need the heavy lifting of a full request
        # run.
        with self.test_request_context(
            path=f'/{view_arg}/nothing/values', base_url='https://localhost/'
        ):
            # explicitly set the view_args, this is a special case
            # for a synthetic test case, usually one would rely on
            # a full request stack to set these.
            flask.request.view_args = {'test_value': view_arg}

            # create dummy ref
            ref = {'id': uuid.uuid4().hex}

            # add the self referential link
            TestResource._add_self_referential_link(
                ref, collection_name='values'
            )

            # Check that the link in fact starts with what we expect
            # including the explicit view arg.
            self.assertTrue(
                ref['links']['self'].startswith(
                    f'https://localhost/v3/{view_arg}'
                )
            )

    def test_json_body_before_req_func_valid_json(self):
        with self.test_request_context(
            headers={'Content-Type': 'application/json'},
            data='{"key": "value"}',
        ):
            # No exception should be raised, everything is happy.
            json_body.json_body_before_request()

    def test_json_body_before_req_func_invalid_json(self):
        with self.test_request_context(
            headers={'Content-Type': 'application/json'}, data='invalid JSON'
        ):
            # keystone.exception.ValidationError should be raised
            self.assertRaises(
                exception.ValidationError, json_body.json_body_before_request
            )

    def test_json_body_before_req_func_no_content_type(self):
        # Unset
        with self.test_request_context(data='{"key": "value"}'):
            # No exception should be raised, everything is happy.
            json_body.json_body_before_request()

        # Explicitly set to ''
        with self.test_request_context(
            headers={'Content-Type': ''}, data='{"key": "value"}'
        ):
            # No exception should be raised, everything is happy.
            json_body.json_body_before_request()

    def test_json_body_before_req_func_unrecognized_content_type(self):
        with self.test_request_context(
            headers={'Content-Type': 'unrecognized/content-type'},
            data='{"key": "value"',
        ):
            # keystone.exception.ValidationError should be raised
            self.assertRaises(
                exception.ValidationError, json_body.json_body_before_request
            )

    def test_json_body_before_req_func_unrecognized_conten_type_no_body(self):
        with self.test_request_context(
            headers={'Content-Type': 'unrecognized/content-type'}
        ):
            # No exception should be raised, everything is happy.
            json_body.json_body_before_request()


class TestKeystoneFlaskUnrouted404(rest.RestfulTestCase):
    def setUp(self):
        super().setUp()
        # unregister the 404 handler we explicitly set in loadapp. This
        # makes the 404 error fallback to a standard werkzeug handling.
        self.public_app.app.error_handler_spec[None].pop(404)

    def test_unrouted_path_is_not_jsonified_404(self):
        with self.test_client() as c:
            path = f'/{uuid.uuid4()}'
            resp = c.get(path, expected_status_code=404)
            # Make sure we're emitting a html error
            self.assertIn('text/html', resp.headers['Content-Type'])
            # Ensure the more generic flask/werkzeug 404 response is emitted
            self.assertTrue(b'404 Not Found' in resp.data)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
import osprofiler.initializer

import keystone.conf

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)


def setup(name, host='0.0.0.0'):  # nosec
    """Setup OSprofiler notifier and enable profiling.

    :param name: name of the service that will be profiled
    :param host: hostname or host IP address that the service will be
                 running on. By default host will be set to 0.0.0.0, but more
                 specified host name / address usage is highly recommended.
    """
    if CONF.profiler.enabled:
        osprofiler.initializer.init_from_conf(
            conf=CONF, context={}, project="keystone", service=name, host=host
        )
        LOG.info(
            "OSProfiler is enabled.\n"
            "Traces provided from the profiler "
            "can only be subscribed to using the same HMAC keys that "
            "are configured in Keystone's configuration file "
            "under the [profiler] section. \n To disable OSprofiler "
            "set in /etc/keystone/keystone.conf:\n"
            "[profiler]\n"
            "enabled=false"
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

DEPRECATED_REASON = (
    "The endpoint groups API is now aware of system scope and default roles."
)

deprecated_list_endpoint_groups = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_endpoint_groups',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_get_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_list_projects_assoc_with_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_projects_associated_with_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_list_endpoints_assoc_with_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_endpoints_associated_with_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_get_endpoint_group_in_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_endpoint_group_in_project',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_list_endpoint_groups_for_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_endpoint_groups_for_project',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_create_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_update_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'update_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_delete_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_add_endpoint_group_to_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'add_endpoint_group_to_project',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_remove_endpoint_group_from_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'remove_endpoint_group_from_project',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)


group_endpoint_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_endpoint_group',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Create endpoint group.',
        operations=[
            {'path': '/v3/OS-EP-FILTER/endpoint_groups', 'method': 'POST'}
        ],
        deprecated_rule=deprecated_create_endpoint_group,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_endpoint_groups',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='List endpoint groups.',
        operations=[
            {'path': '/v3/OS-EP-FILTER/endpoint_groups', 'method': 'GET'}
        ],
        deprecated_rule=deprecated_list_endpoint_groups,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_endpoint_group',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='Get endpoint group.',
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
                ),
                'method': 'GET',
            },
            {
                'path': (
                    '/v3/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
                ),
                'method': 'HEAD',
            },
        ],
        deprecated_rule=deprecated_get_endpoint_group,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'update_endpoint_group',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Update endpoint group.',
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
                ),
                'method': 'PATCH',
            }
        ],
        deprecated_rule=deprecated_update_endpoint_group,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_endpoint_group',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Delete endpoint group.',
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
                ),
                'method': 'DELETE',
            }
        ],
        deprecated_rule=deprecated_delete_endpoint_group,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_projects_associated_with_endpoint_group',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description=(
            'List all projects associated with a specific endpoint group.'
        ),
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/endpoint_groups/'
                    '{endpoint_group_id}/projects'
                ),
                'method': 'GET',
            }
        ],
        deprecated_rule=deprecated_list_projects_assoc_with_endpoint_group,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_endpoints_associated_with_endpoint_group',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='List all endpoints associated with an endpoint group.',
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/endpoint_groups/'
                    '{endpoint_group_id}/endpoints'
                ),
                'method': 'GET',
            }
        ],
        deprecated_rule=deprecated_list_endpoints_assoc_with_endpoint_group,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_endpoint_group_in_project',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description=(
            'Check if an endpoint group is associated with a project.'
        ),
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/endpoint_groups/'
                    '{endpoint_group_id}/projects/{project_id}'
                ),
                'method': 'GET',
            },
            {
                'path': (
                    '/v3/OS-EP-FILTER/endpoint_groups/'
                    '{endpoint_group_id}/projects/{project_id}'
                ),
                'method': 'HEAD',
            },
        ],
        deprecated_rule=deprecated_get_endpoint_group_in_project,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_endpoint_groups_for_project',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='List endpoint groups associated with a specific project.',
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/projects/{project_id}/endpoint_groups'
                ),
                'method': 'GET',
            }
        ],
        deprecated_rule=deprecated_list_endpoint_groups_for_project,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'add_endpoint_group_to_project',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Allow a project to access an endpoint group.',
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/endpoint_groups/'
                    '{endpoint_group_id}/projects/{project_id}'
                ),
                'method': 'PUT',
            }
        ],
        deprecated_rule=deprecated_add_endpoint_group_to_project,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'remove_endpoint_group_from_project',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Remove endpoint group from project.',
        operations=[
            {
                'path': (
                    '/v3/OS-EP-FILTER/endpoint_groups/'
                    '{endpoint_group_id}/projects/{project_id}'
                ),
                'method': 'DELETE',
            }
        ],
        deprecated_rule=deprecated_remove_endpoint_group_from_project,
    ),
]


def list_rules():
    return group_endpoint_policies
